Attribute VB_Name = "modFIR"
'==============================================================
' Descrizione.....: Routines per la sintesi di Filtri FIR e per
'                   la filtrazione corrispondente dei segnali.
'                   Utilita' per le operazioni di "Windowing".
' Nome dei Files..: modFIR.bas
' Data............: 13/06/1999
' Versione........: 1.1 a 32 bits.
' Aggiornamento...: 28/4/2001
' Sistema.........: Visual Basic 6.0 sotto Windows NT 4.0
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.it/flanguasco/
'==============================================================
'
'   Algoritmi tradotti ed adattati dal FORTRAN da:
'     "Digital Filters and their Applications"
'     di V. Cappellini, A. G. Constantinides, P. Emiliani.
'     Remez exchange: pg. 354  364.
'     Window method:  pg. 350.
'
Option Explicit
'
Public Hc#()        ' Vettore dei Coefficienti del Filtro.
Public Ha#()        ' Vettore dei Coefficienti del Filtro arrotondati.
'
Private SRE#()      ' Registro a scorrimento dei Dati da Filtrare.
'
Public WF#()        ' Tabella dei Coefficienti per Windowing.
'
Public Type Window_Type '
    Nome As String      ' Nome della "Window":
    PMin As Double      ' Valore Min. del Parametro associato.
    PMax As Double      ' Valore Max. del Parametro associato.
    PCor As Double      ' Valore corrente del Parametro associato.
End Type
'
'   Variabili e Costanti per "Windows Method".
Private Const A0# = 0.99938     ' Coefficienti per la
Private Const A1# = 0.041186    ' Weber Window.
Private Const A2# = -1.637363   '
Private Const A3# = 0.828217    '
Private Const B0# = 1.496611    '
Private Const B1# = -1.701521   '
Private Const B2# = 0.372793    '
Private Const B3# = 0.0650621   '
'
'   Variabili per "Remez Exchange".
Private NFCNS&, NGRID&, IEXT&()
Private Grid#(), DES#(), WT#()
Private AD#(), ALPHA#(), X#(), Y#()
'
Private Const RemezErr$ = "***** FAILURE TO CONVERGE *****" & vbNewLine & vbNewLine _
    & "Probable cause is the machine Rounding Error." & vbNewLine _
    & "The Impulse Response may be correct." & vbNewLine _
    & "Check the Frequency Response."
Private Const FiltroErr$ = "Le Routines di calcolo hanno trovato" & vbNewLine _
    & "una condizione imprevista." & vbNewLine _
    & "Provare a rivedere i parametri del Filtro." & vbNewLine & vbNewLine
Public Function FilFIR(ByVal dIn#) As Double
'
'   Funzione di filtrazione FIR dei segnali.
'   Devono essere stati definiti, precedentemente, un vettore
'   Hc(0 To NCoeff - 1) contenente gli NCoeff coefficienti del
'   filtro ed un vettore SRE(-1 To NCoeff - 1) da usare come
'   registro a scorrimento dei campioni da filtrare:
'
    Dim K&, dOut#
'
    SRE(-1) = dIn
    For K = UBound(SRE) To 0 Step -1
        SRE(K) = SRE(K - 1)
        dOut = dOut + SRE(K) * Hc(K)
    Next K
'
    FilFIR = dOut
'
'
'
End Function
Private Function I0_Kaiser(ByVal X#, Optional ByRef N&) As Double
'
'   Calcola e ritorna la funzione di Bessel modificata del primo
'   tipo e di ordine zero per 0 <= X <= 20.  N e' il numero
'   di iterazioni impiegate per la convergenza.
'    Da: Theory and Application of Digital Signal Processing.
'        L. R. Rabiner, B. Gold   -   pg. 103
'
    Dim Y#, E#, de#, SDE#
    Const T# = 0.00000001
'
    Y = X / 2#
    E = 1#
    de = 1#
    For N = 1 To 25
        de = de * Y / CDbl(N)
        SDE = de ^ 2
        E = E + SDE
        If E * T - SDE > 0# Then
            I0_Kaiser = E
            Exit Function
        End If
    Next N
'
    N = 0
'
'
'
End Function

Private Function D_D(ByVal K&, ByVal N&, ByVal M&) As Double
'
'
    Dim J&, L&, dd#, Q#
'
    dd = 1#
    Q = X(K)
    For L = 1 To M
        For J = L To N Step M
            If J <> K Then
                dd = 2# * dd * (Q - X(J))
            End If
        Next J
    Next L
    D_D = 1# / dd#
'
'
'
End Function

Private Function WATE(ByVal Tipo$, ByVal TEMP#, ByVal LBAND&, WTX#()) As Double
'
'
    If Tipo$ = "Differentiator" Then
        WATE = WTX(LBAND) / TEMP
    Else
        WATE = WTX(LBAND)
    End If
'
'
'
End Function
Public Function SintesiFIR_RE(ByVal Tipo$, ByVal NFilt&, _
    ByVal LGrid&, ByVal Nbande&, Edge#(), Fx#(), WTX#()) As Boolean
'
'   Sintesi del Filtro con metodo Remez Exchange.
'
    Dim J&, K&, L&, NEG&, NM1&, NOdd&, NZ&, LBAND&
    Dim Change#, DELF#, FUP#, TEMP#, M$
    ReDim Grid#((NFilt + 1) * LGrid / 2)
    ReDim DES#((NFilt + 1) * LGrid / 2), WT#((NFilt + 1) * LGrid / 2)
    ReDim Hc(0 To NFilt - 1)
'
    On Error GoTo SintesiFIR_RE_ERR
'
    NOdd = NFilt Mod 2
    NFCNS = Int(NFilt / 2)
    NEG = IIf(Tipo$ = "Multiple B.Pass/B.Stop", 0, 1)
    If (NOdd = 1) And (NEG = 0) Then NFCNS = NFCNS + 1
'
    ReDim IEXT&(NFCNS + 2), AD#(NFCNS + 2), ALPHA#(NFCNS + 2)
    ReDim X#(NFCNS + 2), Y#(NFCNS + 2)
'
'   SET UP THE DENSE GRID. THE NUMBER OF POINTS IN THE GRID
'   IS (FILTER LENGHT + 1) * GRID DENSITY / 2:
    Grid(1) = Edge(1)
    DELF = CDbl(LGrid * NFCNS)
    DELF = 0.5 / DELF
    If NEG = 1 Then
        If Edge(1) < DELF Then Grid(1) = DELF
    End If
'
'   CALCULATE THE DESIRED MAGNITUDE RESPONSE AND THE WEIGHT
'   FUNCTION ON THE GRID:
    J = 1
    L = 1
    LBAND = 1
    Do
        FUP = Edge(L + 1)
        Do
            TEMP = Grid(J)
            DES(J) = EFF(Tipo$, TEMP, LBAND, Fx())
            WT(J) = WATE(Tipo$, TEMP, LBAND, WTX())
            J = J + 1
            Grid(J) = TEMP + DELF
        Loop While Grid(J) <= FUP
        Grid(J - 1) = FUP
        DES(J - 1) = EFF(Tipo$, TEMP, LBAND, Fx())
        WT(J - 1) = WATE(Tipo$, TEMP, LBAND, WTX())
        LBAND = LBAND + 1
        L = L + 2
        Grid(J) = Edge(L)
    Loop While LBAND <= Nbande
    NGRID = J - 1
    If NEG = NOdd Then
        If Grid(NGRID) > 0.5 - DELF Then NGRID = NGRID - 1
    End If
'
'   SET UP A NEW APPROXIMATION PROBLEM WHICH IS EQUIVALENT
'   TO THE ORIGINAL PROBLEM:
    Select Case NEG
    Case 0
        If NOdd = 0 Then
            For J = 1 To NGRID
                Change = Cos(PI * Grid(J))
                DES(J) = DES(J) / Change
                WT(J) = WT(J) * Change
            Next J
        End If
'
    Case 1
        Select Case NOdd
        Case 0
            For J = 1 To NGRID
                Change = Sin(PI * Grid(J))
                DES(J) = DES(J) / Change
                WT(J) = WT(J) * Change
            Next J
'
        Case 1
            For J = 1 To NGRID
                Change = Sin(PI2 * Grid(J))
                DES(J) = DES(J) / Change
                WT(J) = WT(J) * Change
            Next J
        End Select
    End Select
'
'   INITIAL GUESS FOR THE EXTREMAL FREQUENCIES
'   EQUALLY SPACED ALONG THE GRID:
    TEMP = CDbl(NGRID - 1) / CDbl(NFCNS)
    For J = 1 To NFCNS
        IEXT(J) = (J - 1) * TEMP + 1
    Next J
    IEXT(NFCNS + 1) = NGRID
    NM1 = NFCNS - 1
    NZ = NFCNS + 1
'
'   CALL THE REMEZ EXCHANGE ALGORITHM TO DO
'   THE APPROXIMATION PROBLEM:
    Call Remez(Edge(), Nbande)
'
'   CALCULATE THE IMPULSE RESPONSE:
    Select Case NEG
    Case 0
        Select Case NOdd
        Case 0
            Hc(1 - 1) = 0.25 * ALPHA(NFCNS)
            For J = 2 To NM1
                Hc(J - 1) = 0.25 * (ALPHA(NZ - J) _
                          + ALPHA(NFCNS + 2 - J))
            Next J
            Hc(NFCNS - 1) = 0.5 * ALPHA(1) + 0.25 * ALPHA(2)
'
        Case 1
            For J = 1 To NM1
                Hc(J - 1) = 0.5 * ALPHA(NZ - J)
            Next J
            Hc(NFCNS - 1) = ALPHA(1)
        End Select
'
    Case 1
        Select Case NOdd
        Case 0
            Hc(1 - 1) = 0.25 * ALPHA(NFCNS)
            For J = 2 To NM1
                Hc(J - 1) = 0.25 * (ALPHA(NZ - J) _
                          - ALPHA(NFCNS + 2 - J))
            Next J
            Hc(NFCNS - 1) = 0.5 * ALPHA(1) - 0.25 * ALPHA(2)
               
        Case 1
            Hc(1 - 1) = 0.25 * ALPHA(NFCNS)
            Hc(2 - 1) = 0.25 * ALPHA(NM1)
            For J = 3 To NM1
                Hc(J - 1) = 0.25 * (ALPHA(NZ - J) _
                          - ALPHA(NFCNS + 3 - J))
            Next J
            Hc(NFCNS - 1) = 0.5 * ALPHA(1) - 0.25 * ALPHA(3)
            Hc(NZ - 1) = 0#
        End Select
    End Select
'
    For J = 1 To NFCNS
        K = NFilt + 1 - J
        Hc(K - 1) = Hc(J - 1)
        If NEG = 1 Then Hc(K - 1) = -Hc(K - 1)
    Next J
    If (NEG = 1) And (NOdd = 1) Then Hc(NZ - 1) = 0#
'
'
SintesiFIR_RE_ERR:
    SintesiFIR_RE = (Err = 0)
    If Err <> 0 Then
        M$ = FiltroErr$ & vbNewLine
        M$ = M$ & "Errore " & Str$(Err.Number) & vbNewLine
        M$ = M$ & Err.Description
        MsgBox M$, vbCritical, " Errore in SintesiFIR_RE"
    End If
'
'
'
End Function

Private Sub Remez(Edge#(), ByVal Nbande&)
'
'   Routine di calcolo del "Remez Exchange Algorithm"
'
    Dim J&, JM1&, JP1&, K&, K1&, KN&, KNZ&, KLOW&, KKK&, KUP&
    Dim L&, LUCK&, nu&, NUT&, NUT1&, NM1&
    Dim ITRMAX&, NZ&, NZZ&, NITER&, JET&, JCHNGE&
    Dim DEV#, DEVL#, YNZ#, Y1#, COMP#, DTEMP#, ERRR#, DNUM#, DDEN#
    Dim Aa#, Bb#, GTEMP#, CN#, DELF#, FSH#, Ft#, XT#, XE#
    ReDim A#(NFCNS), P#(NFCNS), Q#(NFCNS)
'
'   THE PROGRAM ALLOWS A MAXIMUM NUMBER OF ITERATIONS OF 25.
'
    ITRMAX = 25
    DEVL = -1#
    NZ = NFCNS + 1
    NZZ = NFCNS + 2
    NITER = 0
    YNZ# = 0#
    Y1 = 0#
'
R100:
    IEXT(NZZ) = NGRID + 1
    NITER = NITER + 1
    If NITER > ITRMAX Then GoTo R400
    For J = 1 To NZ
        DTEMP = Grid(IEXT(J))
        DTEMP = Cos(DTEMP * PI2)
        X(J) = DTEMP
    Next J
    JET = (NFCNS - 1) / 15 + 1
    For J = 1 To NZ
        AD(J) = D_D(J, NZ, JET)
    Next J
    DNUM = 0#
    DDEN = 0#
    K = 1
    For J = 1 To NZ
        L = IEXT(J)
        DTEMP = AD(J) * DES(L)
        DNUM = DNUM + DTEMP
        DTEMP = K * AD(J) / WT(L)
        DDEN = DDEN + DTEMP
        K = -K
    Next J
    DEV = DNUM / DDEN
    nu = 1
    If DEV > 0# Then nu = -1
    DEV = -nu * DEV
    K = nu
    For J = 1 To NZ
        L = IEXT(J)
        DTEMP = K * DEV / WT(L)
        Y(J) = DES(L) + DTEMP
        K = -K
    Next J
'
    If DEV < DEVL Then
        MsgBox RemezErr$, vbCritical, " Errore in Remez"
        GoTo R400
    End If
'
    DEVL = DEV
    JCHNGE = 0
    K1 = IEXT(1)
    KNZ = IEXT(NZ)
    KLOW = 0
    NUT = -nu
    J = 1
'
'   SEARCH FOR THE EXTREMAL FREQUENCIES OF THE BEST APPROXIMATION:
R200:
    If J = NZZ Then YNZ# = COMP
    If J >= NZZ Then GoTo R300
    KUP = IEXT(J + 1)
    L = IEXT(J) + 1
    NUT = -NUT
    If J = 2 Then Y1 = COMP
    COMP = DEV
    If L >= KUP Then GoTo R220
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R220
    COMP = NUT * ERRR
R210:
    L = L + 1
    If L >= KUP Then GoTo R215
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R215
    COMP = NUT * ERRR
    GoTo R210
R215:
    IEXT(J) = L - 1
    J = J + 1
    KLOW = L - 1
    JCHNGE = JCHNGE + 1
    GoTo R200
R220:
    L = L - 1
R225:
    L = L - 1
    If L <= KLOW Then GoTo R250
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP > 0# Then GoTo R230
    If JCHNGE <= 0 Then GoTo R225
    GoTo R260
R230:
    COMP = NUT * ERRR
R235:
    L = L - 1
    If L <= KLOW Then GoTo R240
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R240
    COMP = NUT * ERRR
    GoTo R235
R240:
    KLOW = IEXT(J)
    IEXT(J) = L + 1
    J = J + 1
    JCHNGE = JCHNGE + 1
    GoTo R200
R250:
    L = IEXT(J) + 1
    If JCHNGE > 0 Then GoTo R215
R255:
    L = L + 1
    If L >= KUP Then GoTo R260
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R255
    COMP = NUT * ERRR
    GoTo R210
R260:
    KLOW = IEXT(J)
    J = J + 1
    GoTo R200
R300:
    If J > NZZ Then GoTo R320
    If K1 > IEXT(1) Then K1 = IEXT(1)
    If KNZ < IEXT(NZ) Then KNZ = IEXT(NZ)
    NUT1 = NUT
    NUT = -nu
    L = 0
    KUP = K1
    COMP = YNZ# * 1.00001
    LUCK = 1
R310:
    L = L + 1
    If L >= KUP Then GoTo R315
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R310
    COMP = NUT * ERRR
    J = NZZ
    GoTo R210
R315:
    LUCK = 6
    GoTo R325
R320:   If LUCK > 9 Then GoTo R350
    If COMP > Y1 Then Y1 = COMP
    K1 = IEXT(NZZ)
R325:
    L = NGRID + 1
    KLOW = KNZ
    NUT = -NUT1
    COMP = Y1 * 1.00001
R330:
    L = L - 1
    If L <= KLOW Then GoTo R340
    ERRR = GEE(L, NZ)
    ERRR = (ERRR - DES(L)) * WT(L)
    DTEMP = NUT * ERRR - COMP
    If DTEMP <= 0# Then GoTo R330
    J = NZZ
    COMP = NUT * ERRR
    LUCK = LUCK + 10
    GoTo R235
R340:
    If LUCK = 6 Then GoTo R370
    For J = 1 To NFCNS
        IEXT(NZZ - J) = IEXT(NZ - J)
    Next J
    IEXT(1) = K1
    GoTo R100
R350:
    KN = IEXT(NZZ)
    For J = 1 To NFCNS
        IEXT(J) = IEXT(J + 1)
    Next J
    IEXT(NZ) = KN
    GoTo R100
R370:
    If JCHNGE > 0 Then GoTo R100
'
'   CALCULATION OF THE COEFFICIENTS OF THE BEST APPROXIMATION
'   USING THE INVERSE DISCRETE FOURIER TRANSFORM:
R400:
    NM1 = NFCNS - 1
    GTEMP = Grid(1)
    X(NZZ) = -2#
    CN = 2# * NFCNS - 1#
    DELF = 1# / CN
    KKK = 0
    If (Edge(1) = 0#) And (Edge(2 * Nbande) = 0.5) Then KKK = 1
    If NFCNS <= 3 Then KKK = 1
    If KKK = 0 Then
        DTEMP = Cos(PI2 * Grid(1))
        DNUM = Cos(PI2 * Grid(NGRID))
        Aa = 2# / (DTEMP - DNUM)
        Bb = -(DTEMP + DNUM) / (DTEMP - DNUM)
    End If
    L = 0
    FSH = 0.000001
    For J = 1 To NFCNS
        Ft = (J - 1) * DELF
        XT = Cos(PI2 * Ft)
        If KKK = 0 Then
            XT = (XT - Bb) / Aa
            Ft = Acos(XT) / PI2
        End If
        Do
            L = L + 1
            XE = X(L)
        Loop Until (XT > XE) Or ((XE - XT) < FSH)
        If (XT - XE) < FSH Then
            A(J) = Y(L)
        Else
            Grid(1) = Ft
            A(J) = GEE(1, NZ)
        End If
        If L > 1 Then L = L - 1
    Next J
    Grid(1) = GTEMP
    DDEN = PI2 / CN
    For J = 1 To NFCNS
        DTEMP = 0#
        DNUM = (J - 1) * DDEN
        If NM1 >= 1 Then
            For K = 1 To NM1
                DTEMP = DTEMP + A(K + 1) * Cos(DNUM * K)
            Next K
        End If
        DTEMP = 2# * DTEMP + A(1)
        ALPHA(J) = DTEMP
    Next J
    For J = 2 To NFCNS
        ALPHA(J) = 2# * ALPHA(J) / CN
    Next J
    ALPHA(1) = ALPHA(1) / CN
    If KKK = 0 Then
        P(1) = 2# * ALPHA(NFCNS) * Bb + ALPHA(NM1)
        P(2) = 2# * ALPHA(NFCNS) * Aa
        Q(1) = ALPHA(NFCNS - 2) - ALPHA(NFCNS)
        For J = 2 To NM1
            If J >= NM1 Then
                Aa = 0.5 * Aa
                Bb = 0.5 * Bb
            End If
            P(J + 1) = 0#
            For K = 1 To J
                A(K) = P(K)
                P(K) = 2# * Bb * A(K)
            Next K
            P(2) = P(2) + A(1) * 2# * Aa
            JM1 = J - 1
            For K = 1 To JM1
                P(K) = P(K) + Q(K) + Aa * A(K + 1)
            Next K
            JP1 = J + 1
            For K = 3 To JP1
                P(K) = P(K) + Aa * A(K - 1)
            Next K
            If J <> NM1 Then
                For K = 1 To J
                    Q(K) = -A(K)
                Next K
                Q(1) = Q(1) + ALPHA(NFCNS - 1 - J)
            End If
        Next J
        For J = 1 To NFCNS
            ALPHA(J) = P(J)
        Next J
    End If
    If NFCNS > 3 Then Exit Sub
    ALPHA(NFCNS + 1) = 0#
    ALPHA(NFCNS + 2) = 0#
'
'
'
End Sub

Private Function EFF(ByVal Tipo$, ByVal TEMP#, _
    ByVal LBAND&, Fx#()) As Double
'
'
    If Tipo$ = "Differentiator" Then
        EFF = Fx(LBAND) * TEMP
    Else
        EFF = Fx(LBAND)
    End If
'
'
'
End Function

Private Function GEE(ByVal K&, ByVal N&) As Double
'
'
    Dim J&, P#, C#, DG#, XF#
'
    P = 0#
    XF = Grid(K)
    XF = Cos(PI2 * XF)
    DG = 0#
    For J = 1 To N
        C = XF - X(J)
        C = AD(J) / C
        DG = DG + C
        P = P + C * Y(J)
    Next J
    GEE = P / DG
'
'
'
End Function

Public Function FiltraFIR(S#()) As Double()
'
'   Ritorna il Segnale S#() filtrato usando i Coefficienti Ha#().
'
    Dim I&, K&, LSig_1&, NFILT_1&
    Dim M$, LSig_1O&, All As Boolean
'
    frmMesg.Apri "Filtrazione del Segnale"
    DoEvents
'
'
    LSig_1 = UBound(S)
    NFILT_1 = UBound(Ha)
    ' Mettere allungamento provvisorio di S se LSig_1 < NFilt_1
    If LSig_1 < NFILT_1 Then
        M$ = "Il Segnale e' piu' corto del Filtro." & vbNewLine
        M$ = M$ & "Lo posso allungare aggiungendo zeri in coda?"
        Select Case MsgBox(M$, vbInformation + vbYesNo, " Avvertimento")
        Case vbYes
            LSig_1O = LSig_1
            LSig_1 = NFILT_1
            ReDim Preserve S(0 To NFILT_1)
            All = True
'
        Case vbNo
            ReDim FiltraFIR(0 To LSig_1)   ' Ritorna tutti 0.
            Unload frmMesg
            Exit Function
        End Select
    End If
'
    ReDim F#(0 To LSig_1)       ' Vettore temporaneo dei Dati Filtrati.
    ReDim SRE#(-1 To NFILT_1)   ' Registro a Scorrimento dei
                                ' Dati da Filtrare.
'
    ' Impostazione delle condizioni iniziali del Filtro; tratta
    ' il Segnale di ingresso come periodico con periodo LSig_1 + 1:
    For K = 0 To NFILT_1
        SRE(K) = S(LSig_1 - K)
    Next K
'
    ' Filtrazione per Convoluzione diretta:
    For I = 0 To LSig_1
        F(I) = 0#
        SRE(-1) = S(I)
        For K = NFILT_1 To 0 Step -1
            SRE(K) = SRE(K - 1)
            F(I) = F(I) + SRE(K) * Ha(K)
        Next K
    Next I
    FiltraFIR = F()
'
    If All Then ReDim Preserve S(0 To LSig_1O)
'
    Unload frmMesg
    DoEvents
'
'
'
End Function
Public Function ArrotondamentoFIR(H#(), ByVal ND&) As Double()
'
'   Ritorna i valori arrotondati a ND& decimali del vettore H#().
'   Se ND& < 0 ritorna i valori a Doppia Precisione.
'
    Dim I&, Ft$
    ReDim Tmp#(0 To UBound(H))
'
    If ND < 0 Then
        ArrotondamentoFIR = H()
    Else
        Ft$ = "#0." & String$(ND, "#")
        For I = 0 To UBound(H)
            Tmp(I) = CDbl(Format$(H(I), Ft$))
        Next I
        ArrotondamentoFIR = Tmp()
    End If
'
'
'
End Function


Public Function WinProf(ByVal WNome$, ByVal NWind&, Optional ByVal Par#) As Double()
'
'   Calcola e ritorna un vettore con i coefficienti della Window richiesta.
'    WNome$:    nome della Window richiesta.
'    NWind:     N di coefficienti richiesto.
'    Par:       parametro richiesto per certi tipi di Window.
'
    Dim K&, KK&, NOdd&, NWind1&
    Dim Bb#, Aa#, Ak#, Arg#, ArgL#, ArgW#
    ReDim WF_I#(0 To NWind - 1)
'
    NOdd = NWind Mod 2
    NWind1 = Int(NWind / 2)
    If NOdd = 1 Then WF_I(NWind1) = 1#
'
    Aa = 0.5 * (1# - CDbl(NOdd))
    Bb = CDbl(NWind1) - Aa
'
    ' Calcola la prima meta' del vettore:
    Select Case WNome$
    Case "Blackman"
        For K = 0 To NWind1 - 1
            WF_I(K) = 0.42 - 0.5 * Cos(PI2 * CDbl(K) / CDbl(NWind - 1)) _
                    + 0.08 * Cos(2# * PI2 * CDbl(K) / CDbl(NWind - 1))
        Next K
'
    Case "Hamming generalizzata"
        For K = 0 To NWind1 - 1
            WF_I(K) = Par - (1# - Par) * Cos(PI2 * CDbl(K) / CDbl(NWind - 1))
        Next K
'
    Case "Hamming"
        ' E' l' "Hamming generalizzata" con Par = 0.54:
        For K = 0 To NWind1 - 1
            WF_I(K) = 0.54 - 0.46 * Cos(PI2 * CDbl(K) / CDbl(NWind - 1))
        Next K
'
    Case "Hanning"
        ' E' l' "Hamming generalizzata" con Par = 0.5:
        For K = 0 To NWind1 - 1
            WF_I(K) = 0.5 * (1# - Cos(PI2 * CDbl(K) / CDbl(NWind - 1)))
        Next K
'
    Case "Kaiser"
        For K = 0 To NWind1 - 1
            Ak = Par * Sqr(1# - ((2# * CDbl(K) / CDbl(NWind - 1)) - 1#) ^ 2)
            WF_I(K) = I0_Kaiser(Ak) / I0_Kaiser(Par)
        Next K
'
    Case "Lanczos"
        ArgL = PI / Bb
        For K = 0 To NWind1 - 1
            Ak = CDbl(NWind1 - K) - Aa
            Arg = ArgL * Ak
            WF_I(K) = Abs(Sin(Arg) / Arg) ^ Par
        Next K
'
    Case "Rettangolare"
        For K = 0 To NWind1 - 1
            WF_I(K) = 1#
        Next K
'
    Case "Triangolare"
        For K = 0 To NWind1 - 1
            WF_I(K) = CDbl(K) / CDbl(NWind1)
        Next K
'
    Case "Weber"
        ArgW = 1.5 * Par / Bb
        For K = 0 To NWind1 - 1
            Ak = CDbl(NWind1 - K) - Aa
            Arg = ArgW * Ak
            If (Arg <= 0.75) Then
                WF_I(K) = A0 + A1 * Arg + A2 * Arg ^ 2 + A3 * Arg ^ 3
            Else
                WF_I(K) = B0 + B1 * Arg + B2 * Arg ^ 2 + B3 * Arg ^ 3
            End If
        Next K
    End Select
'
'
    ' Completa il vettore
    ' con la seconda meta':
    For K = 0 To NWind1 - 1
        KK = NWind - 1 - K
        WF_I(KK) = WF_I(K)
    Next K
'
    WinProf = WF_I()
'
'
'
End Function

Public Function SintesiFIR_WM(ByVal Tipo$, ByVal NFilt&, _
    ByVal F1#, ByVal F2#, ByVal WNome$, ByVal Par#) As Boolean
'
'   Sintesi del Filtro con Window Method.
'
    Dim I&, K&, KK&, NOdd&, NFilt1&, M$
    Dim Aa#, Ak#, Af1#, Af2#, Den#
    ReDim Hc(0 To NFilt - 1)
'
    On Error GoTo SintesiFIR_WM_ERR
'
    NOdd = NFilt Mod 2
    NFilt1 = Int(NFilt / 2)
    Aa = 0.5 * (1# - CDbl(NOdd))
'
    WF() = WinProf(WNome$, NFilt, Par)
'
'       Impulse response evaluation
'       ---------------------------
'
    If NOdd = 1 Then
        Select Case Tipo$
        Case "Low Pass"
            Hc(NFilt1) = 2# * F1
        Case "High Pass"
            Hc(NFilt1) = 1# - 2# * F1
        Case "Band Pass"
            Hc(NFilt1) = 2# * (F2 - F1)
        Case "Band Stop"
            Hc(NFilt1) = 1# - 2# * (F2 - F1)
        Case "Differentiator"
            Hc(NFilt1) = 0#
        Case "Hilbert Trans."
            Hc(NFilt1) = 0#
        End Select
    End If
'
    For I = 1 To NFilt1
        K = NFilt1 - I
        Ak = CDbl(I) - Aa
        Af1 = PI2 * Ak * F1
        Af2 = PI2 * Ak * F2
        Den = PI * Ak
        Select Case Tipo$
        Case "Low Pass"
            Hc(K) = Sin(Af1) / Den
        Case "High Pass"
            Hc(K) = (Sin(PI * Ak) - Sin(Af1)) / Den
        Case "Band Pass"
            Hc(K) = (Sin(Af2) - Sin(Af1)) / Den
        Case "Band Stop"
            Hc(K) = (Sin(PI * Ak) - Sin(Af2) + Sin(Af1)) / Den
        Case "Differentiator"
            Hc(K) = -(Sin(Af1) - Af1 * Cos(Af1)) / (Den * Den)
        Case "Hilbert Trans."
            Hc(K) = -(Cos(Af1) - Cos(Af2)) / Den
        End Select
'
        ' Windowing:
        Hc(K) = Hc(K) * WF(K)
    Next I
'
'
    For K = 0 To NFilt1 - 1
        KK = NFilt - 1 - K
        Hc(KK) = Switch(Tipo$ = "Differentiator", -Hc(K), _
                        Tipo$ = "Hilbert Trans.", -Hc(K), True, Hc(K))
    Next K
'
'
SintesiFIR_WM_ERR:
    SintesiFIR_WM = (Err = 0)
    If Err <> 0 Then
        M$ = FiltroErr$ & vbNewLine
        M$ = M$ & "Errore " & Str$(Err.Number) & vbNewLine
        M$ = M$ & Err.Description
        MsgBox M$, vbCritical, " Errore in SintesiFIR_WM"
    End If
'
'
'
End Function


Public Sub RispostaH(H#(), Frq#(), GRs#(), FRs#())
'
'   Calcola la Risposta (Modulo e Fase) nei punti di frequenza
'   del vettore Frq#() di un Filtro FIR dal suo vettore dei
'   Coefficienti H#().
'
'    Frq#():  Vettore delle Frequenze a cui calcolare:
'    GRs#():   il Vettore del Guadagno della Risposta.
'    FRs#():   il Vettore della Fase della Risposta.
'
    Dim I&, K&, NFILT_1
    Dim Omega#, SC#, SS#
'
    NFILT_1 = UBound(H)
'
    For I = 0 To UBound(Frq)
        Omega = PI2 * Frq(I)
        SC = 0#
        SS = 0#
        For K = 0 To NFILT_1
            SC = SC + H(K) * Cos(CDbl(K) * Omega)
            SS = SS - H(K) * Sin(CDbl(K) * Omega)
        Next K
        GRs(I) = Sqr(SC * SC + SS * SS)
        FRs(I) = DATAN2(SS, SC)
    Next I
'
'
'
End Sub

Public Function SintesiFIR_SX(ByVal NFilt&, ByVal CProf&, _
     Frq#(), Fx#(), WTX#(), ByVal Foglio As PictureBox, _
     ByVal FHIn As Boolean) As Boolean
'
'   Sintesi del Filtro con Simplex.
'
    Dim I&, F&, M$
    Dim IBestV&, IWorstV&, ValBV#, ValWV#, ValDf#, NCiclo&, NCC&, OpS$
    ReDim Preserve Hc#(0 To NFilt - 1)
    ReDim HT#(0 To NFilt - 1)
    ReDim GRs#(0 To UBound(Frq)), FRs#(0 To UBound(Frq))
    Dim ValBV_O#, FRinnova As Boolean, FStop As Boolean
'
    Const ErrMax# = 0.000001        ' Valore minimo di arresto sintesi.
    Const MaxCicli& = 20000         ' N. Max di cicli di ricerca minimo.
    Const ValMinDf# = 0.0000001     ' ValDf minimo per il Rinnovo Parametri.
'
    On Error GoTo SintesiFIR_SX_ERR
'
    Foglio.Enabled = False
'
    Call SIMPLES(IBestV, IWorstV, ValBV, ValWV, ValDf, _
                 NFilt, FHIn, Hc(), CProf, Frq(), Fx(), WTX())
    ValBV_O = 1E+39
'
    With frmSimplexOp
        .SimplexOP IBestV, IWorstV, ValBV, ValWV, ValDf, 0, "", FRinnova, FStop
'
        Do
            Call SIMPLEX(IBestV, IWorstV, ValBV, ValWV, ValDf, NCiclo, NCC, OpS$, HT())
            .SimplexOP IBestV, IWorstV, ValBV, ValWV, ValDf, NCiclo, OpS$, FRinnova, FStop
'
            If ValBV_O > ValBV Then
                ValBV_O = ValBV
'
                Hc() = HT()
                RispostaH Hc(), Frq(), GRs(), FRs()
'
                Foglio.Cls
                Foglio.PSet (Frq(0), GRs(0)), vbYellow
                For F = 1 To UBound(Frq)
                    Foglio.Line -(Frq(F), GRs(F)), vbYellow
                Next F
                frmFIR.GraficoCoefficienti Hc()
'
                If IsLoaded(frmLuogoRadici) Then frmFIR.LuogoRadici Hc()
            End If
'
            If FRinnova Or ValDf < ValMinDf Then
                Call SIMPLES(IBestV, IWorstV, ValBV, ValWV, ValDf, _
                             NFilt, False, Hc(), CProf, Frq(), Fx(), WTX())
            End If
'
        Loop While (ValBV > ErrMax) And (NCiclo < MaxCicli) And (Not FStop)
'
'
SintesiFIR_SX_ERR:
    End With
    Unload frmSimplexOp
    If IsLoaded(frmFIR) Then Foglio.Enabled = True
    DoEvents
'
    SintesiFIR_SX = (Err = 0)
    If Err <> 0 Then
        M$ = FiltroErr$ & vbNewLine
        M$ = M$ & "Errore " & Str$(Err.Number) & vbNewLine
        M$ = M$ & Err.Description
        MsgBox M$, vbCritical, " Errore in SintesiFIR_SX"
    End If
'
'
'
End Function
Public Function WinTipi() As Window_Type()
'
'   Imposta i parametri dei tipi di
'   "Window" disponibili:
'    Profilo da File:       WinTipi(-1)
'    Blackman:              WinTipi(0)
'    Hamming generalizzata: WinTipi(1)
'    Hamming:               WinTipi(2)
'    Hanning:               WinTipi(3)
'    Kaiser:                WinTipi(4)
'    Lanczos:               WinTipi(5)
'    Rettangolare:          WinTipi(6)
'    Triangolare:           WinTipi(7)
'    Weber:                 WinTipi(8)
'
    Dim WTipi(-1 To 8) As Window_Type
'
    WTipi(-1).Nome = "--> Nome del File"
    WTipi(-1).PMin = -1
    WTipi(-1).PMax = -1
'
    WTipi(0).Nome = "Blackman"
    WTipi(0).PMin = -1
    WTipi(0).PMax = -1
'
    WTipi(1).Nome = "Hamming generalizzata"
    WTipi(1).PMin = 0.5
    WTipi(1).PMax = 1#
    WTipi(1).PCor = 0.5
'
    WTipi(2).Nome = "Hamming"
    WTipi(2).PMin = -1
    WTipi(2).PMax = -1
'
    WTipi(3).Nome = "Hanning"
    WTipi(3).PMin = -1
    WTipi(3).PMax = -1
'
    WTipi(4).Nome = "Kaiser"
    WTipi(4).PMin = 0.5
    WTipi(4).PMax = 20#
    WTipi(4).PCor = 7.865
'
    WTipi(5).Nome = "Lanczos"
    WTipi(5).PMin = 0.5
    WTipi(5).PMax = 3.5
    WTipi(5).PCor = 1#
'
    WTipi(6).Nome = "Rettangolare"
    WTipi(6).PMin = -1
    WTipi(6).PMax = -1
'
    WTipi(7).Nome = "Triangolare"
    WTipi(7).PMin = -1
    WTipi(7).PMax = -1
'
    WTipi(8).Nome = "Weber"
    WTipi(8).PMin = 0.5
    WTipi(8).PMax = 1#
    WTipi(8).PCor = 1#
'
    WinTipi = WTipi()
'
'
'
End Function
