Attribute VB_Name = "modIIR"
'============================================================
' Descrizione.....: Progetto di Filtri Digitali "IIR" di tipo
'                   Butterworth e Chebyshev per mezzo della
'                   "BILINEAR TRANFORMATION". Routines per la
'                   filtrazione corrispondente dei segnali.
' Nome dei Files..: modIIR.bas
' Data............: 13/06/1999
' Aggiornamento...: 15/05/2001
' Versione........: 1.1 a 32 bits.
' Sistema.........: Visual Basic 6.0 sotto Windows NT 4.0
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.it/flanguasco/
'============================================================
'
'   Gli algoritmi per la sintesi di filtri del tipo Butterworth
'   e Chebyshev sono stati tradotti ed adattati dal FORTRAN da:
'     "Digital Filters and their Applications"
'     di V. Cappellini, A. G. Constantinides, P. Emiliani.
'     pg. 367, 368, 369.
'
'   Per i metodi di calcolo dei filtri elementari vedi Sub Note.
'
Option Explicit
'
Public Ac#()                     ' Coefficienti del Filtro.
Public Bc#()                     ' calcolati in Doppia Precisione.
Public Aa#()                     ' Coefficienti del Filtro.
Public Ba#()                     ' arrotondati a ND Decimali.
'
Private Const NCMax& = 20        ' N. Massimo di Sezioni del Filtro.                                                                         
'
Private W#()                     ' Registri delle Sezioni del Filtro.
'
Private CEB#(1 To 2 * NCMax + 1) ' Vettori in uso durante la Sintesi.
Private AN#(0 To 4, 0 To 2)      '    "        "     "          "
Private FINA#(0 To 2)            '    "        "     "          "
Private FINB#(0 To 2)            '    "        "     "          "
'
Private Const FiltroErr$ = "Le Routines di calcolo hanno trovato" & vbNewLine _
    & "una condizione imprevista." & vbNewLine _
    & "Provare a rivedere i parametri del Filtro." & vbNewLine & vbNewLine

Public Function ArrotondamentoIIR(H#(), ByVal ND&) As Double()
'
'   Ritorna i valori arrotondati a ND& decimali della matrice H#().
'   Se ND& < 0 ritorna i valori a Doppia Precisione.
'
    Dim I&, J&, Ft$
    ReDim Tmp#(0 To UBound(H, 1), 1 To UBound(H, 2))
'
    If ND < 0 Then
        ArrotondamentoIIR = H()
    Else
        Ft$ = "#0." & String$(ND, "#")
        For I = 0 To UBound(H, 1)
            For J = 1 To UBound(H, 2)
                Tmp(I, J) = CDbl(Format$(H(I, J), Ft$))
            Next J
        Next I
        ArrotondamentoIIR = Tmp()
    End If
'
'
'
End Function

Public Function SintesiIIR_BT(ByVal Trans$, ByVal Proty$, _
    ByRef FCt#, ByVal FAt#, ByVal Att#, ByVal Rip#, _
    ByVal F1#, ByVal F2#, ByVal Q#, ByRef AttTot#, ByRef alfa#) As Boolean
'
'   Sintesi del Filtro richiesto con il metodo
'   di Trasformazione Bilineare.
'
    Dim NK&, NCel&
    Dim N&, I&, K&, NC&, M$
    Dim Om#, Cm#, A#, A0#, A1#, A2#, B0#, B1#, B2#      ' Variabili per filtri elementari.
    Dim F#, FF#, FCTA#, FATA#, Eps#, ATL#, res#         ' Variabili per
    Dim ab#, AT#, Bt#, TETAP#, P1#, P2#, P3#            ' filtri di tipo
    Dim CAPPA#, COSTA#, TOT#, cc#, AL#, Rn#             ' Butterworth e Chebyshev.
'
    On Error GoTo SintesiIIR_BT_ERR
'
'-- Filtri elementari ------------------------------------------------------------------
    If Trans$ = "RC Type Low Pass Filter" _
    Or Trans$ = "RC Type High Pass Filter" _
    Or Trans$ = "U(t) = aI(t) + (1-a)U(t-1)" Then
        ' Calcolo di tipi di Filtro elementari del 1 ordine.
        ' Vedi Sub Note:
        Om = PI2 * FCt  ' Pulsazione di taglio [rad/s].
'
        Select Case Trans$
        Case "RC Type Low Pass Filter"
            Cm = Om / Tan(PI * FCt) ' Costante di mappatura.
            A = 1# + Cm / Om
            A0 = 1# / A
            A1 = 1# / A
            B0 = 1#
            B1 = (1# - Cm / Om) / A
'
        Case "RC Type High Pass Filter"
            Cm = Om / Tan(PI * FCt) ' Costante di mappatura.
            A = 1# + Cm / Om
            A0 = (Cm / Om) / A
            A1 = -(Cm / Om) / A
            B0 = 1#
            B1 = (1# - Cm / Om) / A
'
        Case "U(t) = aI(t) + (1-a)U(t-1)"
            ' a := ALFA
            If alfa < 0# Then
                ' E' data FCt; calcola ALFA per avere la FCt voluta:
                alfa = Cos(Om) - 1# + Sqr(Cos(Om) ^ 2 - 4# * Cos(Om) + 3#)
            ElseIf alfa < 2# * (Sqr2 - 1#) Then
                ' E' data ALFA; calcola la FCt (se < 0.5) corrispondente:
                Om = Acos(((alfa ^ 2) + 2# * alfa - 2#) / (2# * alfa - 2#))
                FCt = Om / PI2
            End If
            A0 = alfa
            A1 = 0#
            B0 = 1#
            B1 = -(1# - alfa)
        End Select
'
        NCel = 1
        NK = 1
        ReDim Ac#(0 To NK, 1 To NCel)   ' Coefficienti del Filtro
        ReDim Bc#(0 To NK, 1 To NCel)   ' calcolati in Doppia Precisione.
        Ac(0, 1) = A0
        Ac(1, 1) = A1
        Bc(0, 1) = B0
        Bc(1, 1) = B1
        AttTot = 1#
'
        GoTo SintesiIIR_BT_End
'
    ElseIf Trans$ = "Resonator" _
        Or Trans$ = "Notch Filter" Then
        ' Calcolo di tipi di Filtro elementari del 2 ordine.
        ' Vedi Sub Note:
        Om = PI2 * FCt          ' Pulsazione di risonanza [rad/s].
        Cm = Om / Tan(PI * FCt) ' Costante di mappatura.
'
        ' Calcola i poli:
        A = 1# + Cm / (Q * Om) + (Cm ^ 2 / Om ^ 2)
        B0 = 1#
        B1 = 2# * (1# - Cm ^ 2 / Om ^ 2) / A
        B2 = (1# - Cm / (Q * Om) + (Cm ^ 2 / Om ^ 2)) / A
'
        Select Case Trans$
        Case "Resonator"
            ' Mette due zeri reali a  1:
            A0 = Cm / A
            A1 = 0#
            A2 = -Cm / A
'
        Case "Notch Filter"
            ' Mette una coppia di zeri complessi coniugati:
            Rn = Sqr(B2)
                            ' TETAP = Acos(-B1 / (2# * Rn))
            A0 = 1#
            A1 = B1 / Rn    ' -2 * Cos(TETAP)
            A2 = 1#         ' Cos(TETAP) ^ 2 + Sin(TETAP) ^ 2
        End Select
'
        NCel = 1
        NK = 2
        ReDim Ac#(0 To NK, 1 To NCel)   ' Coefficienti del Filtro
        ReDim Bc#(0 To NK, 1 To NCel)   ' calcolati in Doppia Precisione.
        Ac(0, 1) = A0
        Ac(1, 1) = A1
        Ac(2, 1) = A2
        Bc(0, 1) = B0
        Bc(1, 1) = B1
        Bc(2, 1) = B2
        AttTot = 1#
'
        GoTo SintesiIIR_BT_End
    End If
'
'-- Filtri Butterworth e Chebyshev -----------------------------------------------------
    FCTA = Tan(PI * FCt)
    FATA = Tan(PI * FAt)
    FF = FATA / FCTA
    If Proty$ = "Chebyshev" Then Eps = Sqr(1# / (1# - Rip / 100#) ^ 2# - 1#)
    ATL = 10# ^ (Att / 10#)
'
'       COMPUTATION OF THE ORDER OF THE FILTER.
'       ---------------------------------------
'
    N = 0
    Do
        N = N + 1
        Select Case Proty$
        Case "Butterworth"
            res# = 1# / (1# + FF ^ (2 * N))
'
        Case "Chebyshev"
            CEB(1) = 1#
            CEB(2) = FF
            If N > 1 Then
                   F = FF + FF
                   For I = 2 To N
                       CEB(I + 1) = F * CEB(I) - CEB#(I - 1)
                   Next I
            End If
            res = 1# / (1# + (Eps * CEB(N + 1)) ^ 2)
        End Select
    Loop While res >= ATL
'
    If N Mod 2 = 1 Then
        N = N + 1
        Select Case Proty$
        Case "Butterworth"
            res = 1# / (1# + FF ^ (2 * N))
'
        Case "Chebyshev"
            CEB(1) = 1
            CEB(2) = FF
            F = FF + FF
            For I = 2 To N
                CEB(I + 1) = F * CEB(I) - CEB(I - 1)
            Next I
            res = 1# / (1# + (Eps * CEB(N + 1)) ^ 2)
        End Select
    End If
'
    If N > 2 * NCMax Then
        M$ = "**** ERRORE ****" & vbNewLine
        M$ = M$ & vbNewLine & "Per ottenere il Filtro richiesto" & vbNewLine
        M$ = M$ & "si devono usare " & Str$(N / 2) & " Sezioni." & vbNewLine
        M$ = M$ & "Il numero Massimo e' " & Str$(NCMax) & "." & vbNewLine
        M$ = M$ & vbNewLine & "Provare a variare i dati del Filtro."
        MsgBox M$, vbCritical, " Sintesi IIR"
        SintesiIIR_BT = False
        Unload frmMesg
        DoEvents
        Exit Function
    End If
'
    AttTot = 10# * Log(res) / Log10
    NCel = N / 2
    NK = Switch(Trans$ = "No Transformation", 2, _
                Trans$ = "Low Pass -> Low Pass", 2, _
                Trans$ = "Low Pass -> High Pass", 2, _
                Trans$ = "Low Pass -> Band Pass", 4, _
                Trans$ = "Low Pass -> Band Stop", 4)
    ReDim Ac#(0 To NK, 1 To NCel)    ' Coefficienti del Filtro.
    ReDim Bc#(0 To NK, 1 To NCel)    ' Calcolati in Doppia Precisione.
    ReDim POLRE#(1 To NCel)          ' Vettori in uso durante la Sintesi.
    ReDim POLIM#(1 To NCel)          '    "        "     "          "
'
'       POLE POSITION EVALUATION.
'       -------------------------
'
    AT = FCTA
    Bt = FCTA
    If Proty$ = "Chebyshev" Then
        ab = ((1# + Sqr(1# + Eps ^ 2)) / Eps) ^ (1# / N)
        AT = (ab - 1# / ab) * AT / 2#
        Bt = (ab + 1# / ab) * Bt / 2#
    End If
    For NC = 1 To NCel
        TETAP = PI# * (2# * NC - 1#) / (2# * N)
        POLRE(NC) = -AT * Cos(TETAP)
        POLIM(NC) = Bt * Sin(TETAP)
    Next NC
    For NC = 1 To NCel
        P1 = POLIM(NC) ^ 2
        P2 = (1# - POLRE(NC)) ^ 2
        P3 = POLRE(NC) ^ 2
        POLRE(NC) = (1# - P3 - P1) / (P2 + P1)
        POLIM(NC) = (2# * POLIM(NC)) / (P2 + P1)
    Next NC
'
'       EVALUATION OF THE SECOND ORDER SECTION COEFFICIENTS.
'       ----------------------------------------------------
'
    For NC = 1 To NCel
        Bc(0, NC) = 1#
        Bc(1, NC) = -2# * POLRE(NC)
        Bc(2, NC) = POLRE(NC) * POLRE(NC) + POLIM(NC) * POLIM(NC)
    Next NC
'
    Select Case Proty$
    Case "Butterworth"
        COSTA = 1#
'
    Case "Chebyshev"
        COSTA = (1# - Rip / 100#) ^ (1# / NCel)
    End Select
'
    For NC = 1 To NCel
        TOT = 4# / (Bc(0, NC) + Bc(1, NC) + Bc(2, NC))
        cc = TOT / COSTA
        Ac(0, NC) = 1# / cc
        Ac(1, NC) = 2# / cc
        Ac(2, NC) = 1# / cc
    Next NC
'
'       FREQUENCY TRANSFORMATION.
'       -------------------------
'
    Select Case Trans$
'
    Case "No Transformation"
        GoTo SintesiIIR_BT_End
'
        Case "Low Pass -> Low Pass"
        AL = Sin(PI * (FCt - F1)) / Sin(PI * (FCt + F1))
        A0 = -AL
        A1 = 1#
        A2 = 0#
        B0 = A1
        B1 = A0
        B2 = 0#
'
    Case "Low Pass -> High Pass"
        AL = -Cos(PI * (FCt + F1)) / Cos(PI * (FCt - F1))
        A0 = -AL
        A1 = -1#
        A2 = 0#
        B0 = -A1
        B1 = -A0
        B2 = 0#
'
    Case "Low Pass -> Band Pass"
        AL = Cos(PI * (F2 + F1)) / Cos(PI * (F2 - F1))
        CAPPA = Tan(PI * FCt) / Tan(PI * (F2 - F1))
        A0 = -(CAPPA - 1#) / (CAPPA + 1#)
        A1 = 2# * AL * CAPPA# / (CAPPA + 1#)
        A2 = -1#
        B0 = -A2
        B1 = -A1
        B2 = -A0
'
    Case "Low Pass -> Band Stop"
        AL = Cos(PI * (F2 + F1)) / Cos(PI * (F2 - F1))
        CAPPA = Tan(PI * FCt) * Tan(PI * (F2 - F1))
        A0 = (1# - CAPPA) / (1# + CAPPA)
        A1 = -2# * AL / (1# + CAPPA)
        A2 = 1#
        B0 = A2
        B1 = A1
        B2 = A0
    End Select
'
    AN(0, 0) = B0 * B0
    AN(1, 0) = 2# * B0 * B1
    AN(2, 0) = B1 * B1 + 2# * B0 * B2
    AN(3, 0) = 2# * B1 * B2
    AN(4, 0) = B2 * B2
    AN(0, 1) = A0 * B0
    AN(1, 1) = A0 * B1 + A1 * B0
    AN(2, 1) = A0 * B2 + A1 * B1 + A2 * B0
    AN(3, 1) = A1 * B2 + A2 * B1
    AN(4, 1) = A2 * B2
    AN(0, 2) = A0 * A0
    AN(1, 2) = 2# * A0 * A1
    AN(2, 2) = A1 * A1 + 2# * A0 * A2
    AN(3, 2) = 2# * A1 * A2
    AN(4, 2) = A2 * A2
'
    For NC = 1 To NCel
        For K = 0 To 2
            FINA(K) = Ac(K, NC)
            FINB(K) = Bc(K, NC)
        Next K
        For K = 0 To NK
            Ac(K, NC) = 0#
            Bc(K, NC) = 0#
            For I = 0 To 2
                Ac(K, NC) = AN(K, I) * FINA(I) + Ac(K, NC)
                Bc(K, NC) = AN(K, I) * FINB(I) + Bc(K, NC)
            Next I
        Next K
    Next NC
    For NC = 1 To NCel
        Rn = Bc(0, NC)
        For K = 0 To NK
            Ac(K, NC) = Ac(K, NC) / Rn
            Bc(K, NC) = Bc(K, NC) / Rn
        Next K
    Next NC
'
'
SintesiIIR_BT_End:
SintesiIIR_BT_ERR:
    SintesiIIR_BT = (Err = 0)
    If Err <> 0 Then
        M$ = FiltroErr$ & vbNewLine
        M$ = M$ & "Errore " & Str$(Err.Number) & vbNewLine
        M$ = M$ & Err.Description
        MsgBox M$, vbCritical, " Errore in SintesiIIR_BT"
    End If
'
'
'
End Function

Public Function FiltraIIR(S#()) As Double()
'
'   Ritorna il Segnale S#() filtrato usando i Coefficienti Aa#(), Ba#().
'
    Dim NK&, NCel&
    Dim I&, K&, LSig_1&
'
    frmMesg.Apri "Filtrazione del Segnale"
    DoEvents
'
    LSig_1 = UBound(S)
    ReDim F#(0 To LSig_1)       ' Vettore temporaneo dei Dati Filtrati.
'
    NK = UBound(Ac, 1)
    NCel = UBound(Ac, 2)
    ReDim W(0 To NK, 1 To NCel)
'
    ' Imposta i valori iniziali nella matrice W(); tratta il
    ' Segnale di ingresso come periodico con periodo LSig_1 + 1:
    For I = 0 To LSig_1
        F(I) = FilIIR(S(I), NK, NCel)
    Next I
'
    For I = 0 To LSig_1
        F(I) = FilIIR(S(I), NK, NCel)
    Next I
    FiltraIIR = F()
'
    Unload frmMesg
    DoEvents
'
'
'
End Function

Private Function FilIIR(ByVal vi#, ByVal NK&, ByVal NCel&) As Double
'
'   Filtrazione IIR del Segnale:
'
    Dim C&, K&, Y#
'
    Y = vi
    For C = 1 To NCel
        W(0, C) = Y
        Y = 0#
        For K = NK To 1 Step -1
            W(0, C) = W(0, C) - Ba(K, C) * W(K, C)
            Y = Y + Aa(K, C) * W(K, C)
            W(K, C) = W(K - 1, C)
        Next K
        Y = Y + Aa(0, C) * W(0, C)
    Next C
'
    FilIIR = Y
'
'
'
End Function
Private Sub Note()
'
'   Metodi di calcolo dei filtri elementari.
'   - Algoritmi di trasformazione da "Digital Signal Processing"
'     di W. D. Stanley - pg. 172, 173, 174.
'   - La funzione di trasferimento del risuonatore reale parallelo
'     e' tratta da "Teoria delle Reti Elettriche", Appunti dai
'     corsi del Politecnico di Torino - pg. (1.3)1 e seg.
'   - L' idea della sostituzione degli zeri per il "Notch Filter"
'     proviene da: http://www-users.cs.york.ac.uk/~fisher/mkfilter/res.html
'
'---------------------------------------------------------------------------------------
'   "RC Type Low Pass Filter":
'   La funzione di trasferimento e':
'
'   U(p)       1                    con:
'   ---- = -----------               Fc = 1/(2*PI*RC)   frequenza di taglio a -3 [dB] [Hz].
'   I(p)   1 + RC*p                  C, R               capacita' e resistenza
'                                                       del circuito.
'
'   Le formule di trasformazione (1) da:
'
'           A0 + A1*p                 1
'    G(p) = ---------   -->  -------------------
'           B0 + B1*p        1 + (1/(2*PI*Fc))*p
'   a:
'                    -1
'           a0 + a1*z
'    H(z) = -----------
'                    -1
'           1  + b1*z
'
'   sono:                           con:
'    Cm = 2*PI*FCt/Tan(2*PI*FCt/2)   FCt = Fc/Fs  -->  0 <= FCt <= 0.5
'    A  =  B0 + B1*Cm                A0 = 1
'    a0 = (A0 + A1*Cm)/A             A1 = 0
'    a1 = (A0 - A1*Cm)/A             B0 = 1
'    b1 = (B0 - B1*Cm)/A             B1 = 1/(2*PI*FCt)
'
'
'---------------------------------------------------------------------------------------
'   "RC Type High Pass Filter":
'   La funzione di trasferimento e':
'
'   U(p)      RCp                   con:
'   ---- = -----------               Fc = 1/(2*PI*RC)   frequenza di taglio a -3 [dB] [Hz].
'   I(p)   1 + RC*p                  C, R               capacita' e resistenza
'                                                       del circuito.
'
'   si usano le formule di trasformazione (1) con:
'                                              FCt = Fc/Fs  -->  0 <= FCt <= 0.5
'                                              A0 = 0
'                                              A1 = 1/(2*PI*FCt)
'                                              B0 = 1
'                                              B1 = 1/(2*PI*FCt)
'
'
'---------------------------------------------------------------------------------------
'   "Resonator":
'   La funzione di trasferimento e':
'
'    U(p)                  P                    con:
'    ---- = ---------------------------------    Oc = 1/Sqr(L*C)    pulsazione di risonanza [rad/s].
'    I(P)   1 + (1/(Q*Oc))*P + (1/(Oc^2))*P^2    Fc = Oc/(2*PI)     frequenza di risonanza [Hz].
'                                                L, C, Rp           induttanza, capacita' e
'                                                                   resistenza in parallelo
'                                                                   del circuito risonante.
'                                                Q = Oc*Rp*C        coefficiente di risonanza.
'
'   Le formule di trasformazione (2) da:
'
'           A0 + A1*p + A2*p^2                         p
'    G(p) = ------------------   -->  ---------------------------------
'           B0 + B1*p + B2*p^2        1 + (1/(Q*Oc))*P + (1/(Oc^2))*P^2
'   a:
'                    -1     -2
'           a0 + a1*z + a2*z
'    H(z) = ------------------
'                    -1     -2
'           1  + b1*z + b2*z
'
'   sono:                             con:
'    Cm = 2*PI*FCt/Tan(2*PI*FCt/2)     FCt = Fc/Fs  -->  0 <= FCt <= 0.5
'    A  =  B0 + B1*Cm + B2*Cm^2        A0 = 0
'    a0 = (A0 + A1*Cm + A2*Cm^2)/A     A1 = 1
'    a1 = (2*A0 - 2*A2*Cm^2)/A         A2 = 0
'    a2 = (A0 - A1*Cm + A2*Cm^2)/A     B0 = 1
'    b1 = (2*B0 - 2*B2*Cm^2)/A         B1 = 1/(2*PI*FCt*Q)
'    b2 = (B0 - B1*Cm + B2*Cm^2)/A     B2 = 1/(2*PI*FCt)^2
'
'---------------------------------------------------------------------------------------
'   "Notch Filter":
'   L' algoritmo precedente costruisce una funzione di trasferimento H(z) avente
'   una coppia di zeri reali a  1; questo consente di avere guadagno zero alle
'   basse ed alle alte frequenze.
'   Il "Notch Filter" viene ricavato dal "Resonator" sostituendo questa coppia di
'   zeri reali con due zeri complessi coniugati, giacenti sul cerchio unitario e
'   con argomento uguale a quello dei poli (anch' essi complessi coniugati).
'
'   Se Zp = r*Exp( i*Theta) sono i poli di H(z), gli zeri devono essere:
'
'    Zz = Exp( i*Theta)
'
'   Viene usata la relazione:
'
'                          -1                        -1
'    (1 - r*Exp(+i*Theta)*z  )*(1 - r*Exp(-i*Theta)*z  ) =
'
'                          -1   2   -2           -1      -2
'    = 1 - 2*r*Cos(Theta)*z  + r * z   = 1 + b1*z  + b2*z       (i = Sqr(-1))
'
'   da cui:
'
'    r = Sqr(b2)
'    Theta = Acos(-b1/(2*Sqr(b2))
'
'   e quindi:
'
'             -1     -2                       -1  -2
'    a0 + a1*z + a2*z  = 1 + (b1/(2*Sqr(b2))*z + z
'
'---------------------------------------------------------------------------------------
'   "U(t) = aI(t) + (1-a)U(t-1)"
'   La funzione di trasferimento e':
'
'                a
'    H(z) = ----------------
'                         -1
'           1  - (1 - a)*z
'
'   Usando, invertite, le formule di trasformazione (1) si ha:
'
'              a*Cm - a*p
'    G(p) = ---------------             con: Cm = Oc/Tan(Oc/2)
'           a*Cm + (2 - a)*p                 Oc = 2*PI*Fc
'
'   La pulsazione di taglio Oc si ha per:
'
'          |a*Cm - a*i*Oc|       1
'    ----------------------- = ------   con: p = i*Oc
'      |a*Cm + (2 - a)*i*Oc|   Sqr(2)        i = Sqr(-1)
'
'   da cui si ricava il valore di a necessario per
'   ottenere la frequenza di taglio FCt voluta:
'
'    a = Cos(Oc) - 1 + Sqr(Cos(Oc)^2 - 4*Cos(Oc) + 3)   sempre con: Oc  = 2*PI*FCt
'                                                                   FCt = Fc/Fs
'
'   Se invece viene dato a, si ottiene una FCt corrispondente:
'
'    FCt = Acos(((a^2) + 2*a - 2)/(2*a - 2))/(2*PI)     solo per:   a   < 2*(Sqr(2) - 1)
'                                                       e:          FCt < 0.5
'
'
'
End Sub
