Attribute VB_Name = "modSigPro"
'=============================================================
' Descrizione.....: Routines di calcolo di DFT, FFT (diretta
'                   ed inversa), FHT, delle funzioni di Auto
'                   e Mutua correlazione, della Media Mobile
'                   e dello Spettro Mobile (sliding spectrum).
' Nome dei Files..: modSigPro.bas, modComplex.bas
' Data............: 13/06/1999
' Aggiornamento...: 31/03/2001
' Aggiornamento...: 13/05/2001 (funzioni di correlazione)
' Aggiornamento...: 07/04/2004 (spettro mobile)
' Versione........: 1.2 a 32 bits.
' Sistema.........: Visual Basic 6.0 sotto Windows NT 4.0
' Scritto da......: F. Languasco 
' E-Mail..........: MC7061@mclink.it
' DownLoads a.....: http://www.flanguasco.org
'                   http://members.xoom.it/flanguasco/
'=============================================================
'
Option Explicit
'
' ---- Spettro mobile (sliding spectrum) --------------------------------------
'
'   Implementazione dei circuiti tratti da:
'    "Theory and Application of Digital Signal Processing"
'    di L. Rabiner e B. Gold. - pg. 382, 383.
'
Private NSm&                    ' N di campioni per il calcolo dello spettro mobile.
Private z1() As Complex         ' Vettore dei coefficienti di calcolo con convoluzione
                                ' diretta dello spettro mobile.
Private SmRE#()                 ' Registro a scorrimento dei campioni del segnale.
Private z1_1 As Complex         ' Coefficienti per il
Private z1_N As Complex         ' calcolo ricorsivo.
Private Sn_1 As Complex         ' Sn * Z^-1.
Private SmNorm#                 ' Fattore di normalizzazione sul N di campioni.
Private Ic&, Kc&, NSm1&         ' Indici in SmRE() per buffer circolare.
Public Function SPM(ByVal Xn#) As Double
'
'   Funzione per il calcolo con convoluzione diretta dello
'   spettro mobile (sliding spectrum) su NsM campioni.
'   Devono essere stati definiti, precedentemente, il vettore
'   dei coefficienti z1() ed un vettore SmRE(0 To NsM - 1) da
'   usare come registro a scorrimento dei campioni del segnale.
'
    Dim K&, Sn As Complex
'
    Sn.re = Xn
    For K = NSm - 1 To 1 Step -1
        ' Scorrimento del segnale:
        SmRE(K) = SmRE(K - 1)
'
        ' Convoluzione:
        'Sn = CSom(Sn, CMol(CCmp(SmRE(K), 0#), Z1(K)))
        Sn.re = Sn.re + SmRE(K) * z1(K).re
        Sn.im = Sn.im + SmRE(K) * z1(K).im
    Next K
    SmRE(0) = Xn
'
    SPM = CAbs(Sn) / SmNorm
'
'
'
End Function
Public Function SPM_R(ByVal Xn#) As Double
'
'   Funzione per il calcolo ricorsivo dello spettro mobile
'   (sliding spectrum) su NsM campioni.
'   Devono essere stati definiti, precedentemente, i coefficienti
'   z1_1, z1_N ed un vettore SmRE(0 To NsM) da usare come
'   registro a scorrimento dei campioni del segnale.
'
    Dim K&, Xz1_N As Complex, Yz1_1 As Complex, SnI As Complex, Sn As Complex
'
    ' SmRE() usato come registro a scorrimento:
'    For K = NSm To 1 Step -1
'        SmRE(K) = SmRE(K - 1)
'    Next K
'    SmRE(0) = CDbl(Xn)
'
'    'XZ1_N = CMol(CCmp(SmRE(NSm), 0#), Z1_N)
'    XZ1_N.re = SmRE(NSm) * z1_N.re
'    XZ1_N.im = SmRE(NSm) * z1_N.im
'    'SnI = CDif(CCmp(SmRE(0), 0#), XZ1_N)
'    SnI.re = SmRE(0) - XZ1_N.re
'    SnI.im = -XZ1_N.im
'
    ' SmRE() usato come buffer circolare:
    Ic = ((Ic + 1) Mod NSm1)
    Kc = ((Ic + 1) Mod NSm1)
    SmRE(Ic) = CDbl(Xn)
'
    Xz1_N.re = SmRE(Kc) * z1_N.re
    Xz1_N.im = SmRE(Kc) * z1_N.im
    SnI.re = SmRE(Ic) - Xz1_N.re
    SnI.im = -Xz1_N.im
'
    Yz1_1 = CMol(Sn_1, z1_1)
    Sn = CSom(SnI, Yz1_1)
    Sn_1 = Sn
'
    SPM_R = CAbs(Sn) / SmNorm
'
'
'
End Function
Public Sub SPM_Init(ByVal NSm_I&, ByVal Freq#)
'
'   Calcolo dei coefficienti necessari alle
'   due implementazioni dello spettro mobile:
'
    Dim I&, OmegaT#
'
    NSm = NSm_I             ' N di campioni per il calcolo dello spettro mobile.
    NSm1 = NSm + 1
    ReDim z1(0 To NSm)      ' Vettore dei coefficienti di calcolo con
                            ' convoluzione diretta dello spettro mobile;
                            ' z1(0) = e^j*omega.
    ReDim SmRE#(0 To NSm)   ' Registro a scorrimento dei campioni del segnale.
    SmNorm = CDbl(NSm) / 2# ' Fattore di normalizzazione sul N di campioni.
'
    'OmegaT = PI2 * Freq * TS    ' Per frequenza in [Hz].
    OmegaT = PI2 * Freq         ' Per frequenza in [f/fs].
    z1(0) = CExp(CCmp(0#, OmegaT))
'
    ' Per calcolo con convoluzione diretta:
    For I = 1 To NSm - 1
        z1(I) = CPtN(z1(0), -I)
    Next I
'
    ' Per calcolo ricorsivo:
    Sn_1 = CCmp(0#, 0#)
    z1_1 = CPtN(z1(0), -1)
    z1_N = CPtN(z1(0), -NSm)
'
'
'
End Sub
Sub FHT(D() As Double, R() As Double, X() As Double, _
    S() As Double, F() As Double, ByRef NVAL&, ByRef NFRE&)
'
'   Definizioni:
'
'   FHT = Fast Hartley Transform.
'
'   Entra con:
'       D()   = Vettore dei dati Reali del Segnale(t).
'       NVAL  = Numero max. dei dati in D() da usare.
'
'   Esce  con:
'       R()   = Vettore dei valori Reali      della Trasformata(f).
'       X()   = Vettore dei valori Immaginari della Trasformata(f).
'       S()   = Vettore dello Spettro di Amp. della Trasformata(f).
'       F()   = Vettore delle Fasi            della Trasformata(f) [Radianti].
'       NVAL  = Numero  dei dati usati        dalla Trasformata(f).
'       NFRE  = Numero  dei valori di Frequenza calcolati.
'
'   I valori della Trasformata in R(), X() NON vengono
'   normalizzati sulla lunghezza del Segnale 2^M.
'
'   I vettori R(), X(), S() e F() sono dimensionati in questa routine.
'   Gli NN dati nei vettori sono organizzati come Vettore(0 To NN - 1).
'
'   Viene calcolata in R(0 To NFRE), X(0 To NFRE) solo la prima
'   meta' dello spettro per i valori di frequenza  NFRE = NVAL/2
'   piu' la Componente Continua in R(0).
'
'   Tratto da: "Faster than Fast Fourier" di M. O'Neill, Byte April 88.
'
    Dim M&, NDV4&, NMN1&, I&, J&, K&, L&, N&
    Dim I0&, I1&, I2&, I3&, N1&, K1&
    Dim W#, A#, T#, Ti#, Td#, V#, z#
'
    M = Int(Log(CDbl(NVAL) + 0.5) / Log2)
    NVAL = 2 ^ M
    NFRE = NVAL / 2
    NDV4 = NVAL / 4
    NMN1 = NVAL - 1
'
    ReDim R(0 To NFRE), X(0 To NFRE)
    ReDim S(0 To NFRE), F(0 To NFRE)
'
    ReDim B#(0 To NVAL / 4)
    ReDim C#(0 To NVAL / 4)
    ReDim HT#(0 To NVAL - 1)
'
    For I = 0 To NVAL - 1
        HT(I) = D(I)
    Next I
'
    W = PI2 / NVAL
    A = 0#
    B(0) = 0#
    C(0) = 1#
    For I = 1 To NVAL / 4
        A = A + W
        B(I) = Sin(A)
        C(I) = Cos(A)
    Next I
'
    J = 0
    For I = 0 To NMN1 - 1
        If (I < J) Then Swap HT(J), HT(I)
        K = NFRE
        Do While (K - 1 < J)
            J = J - K
            K = K / 2
        Loop
        J = J + K
    Next I
'
    For I = 0 To NMN1 Step 2
        T = HT(I)
        HT(I) = T + HT(I + 1)
        HT(I + 1) = T - HT(I + 1)
    Next I
    For I = 0 To NMN1 Step 4
        T = HT(I)
        HT(I) = T + HT(I + 2)
        HT(I + 2) = T - HT(I + 2)
        T = HT(I + 1)
        HT(I + 1) = T + HT(I + 3)
        HT(I + 3) = T - HT(I + 3)
    Next I
'
    N = M - 1
    I1 = 4
    For I3 = 3 To M
        I2 = I1 + I1
        N = N - 1
        I0 = 2 ^ (N - 1)
        For N1 = 0 To NMN1 Step I2
            I = N1
            L = I + I1
            Ti = HT(I)
            HT(I) = Ti + HT(L)
            HT(L) = Ti - HT(L)
            K = L - 1
            J = I0
            Do While J <= NDV4
                I = I + 1
                L = I + I1
                K1 = K + I1
                V = HT(L) * C(J) + HT(K1) * B(J)
                z = HT(L) * B(J) - HT(K1) * C(J)
                Ti = HT(K)
                Td = HT(I)
                HT(K1) = Ti - z
                HT(K) = Ti + z
                HT(L) = Td - V
                HT(I) = Td + V
                K = K - 1
                J = J + I0
            Loop
            K1 = K + I1
        Next N1
        I1 = I2
    Next I3
'
    ' Calcolo della parte Reale, della parte Immaginaria
    ' e dello Spettro di Ampiezza e delle Fasi:
    R(0) = HT(0)
    X(0) = 0#
    S(0) = Abs(R(0)) / CDbl(NVAL)
    F(0) = IIf(R(0) < 0#, PI, 0#)
    For J = 1 To NFRE - 1
        R(J) = HT(J) + HT(NVAL - J)
        X(J) = HT(J) - HT(NVAL - J)
        S(J) = Sqr(R(J) * R(J) + X(J) * X(J)) / CDbl(NVAL)
        F(J) = DATAN2(-X(J), R(J))
    Next J
    R(NFRE) = HT(NFRE)
    X(NFRE) = 0#
    S(NFRE) = Abs(R(NFRE)) / CDbl(NVAL)
    F(NFRE) = IIf(R(NFRE) < 0#, PI, 0#)
'
'
'
End Sub
Sub DFT(D() As Double, R() As Double, X() As Double, _
    S() As Double, F() As Double, ByVal NVAL&, ByRef NFRE&, _
    ByVal prbAvanz As ProgressBar)
'
'   Definizioni:
'
'   DFT = Discrete Fourier Transform.
'
'   Entra con:
'       D()    = Vettore dei dati Reali del Segnale(t).
'       NVAL   = Numero max. dei dati in D() da usare.
'
'   Esce  con:
'       R()    = Vettore dei valori Reali      della Trasformata(f).
'       X()    = Vettore dei valori Immaginari della Trasformata(f).
'       S()    = Vettore dello Spettro di Amp. della Trasformata(f).
'       F()    = Vettore delle Fasi            della Trasformata(f) [Radianti].
'       NFRE   = Numero  dei valori di Frequenza calcolati.
'
'   I valori della Trasformata in R(), X() NON vengono
'   normalizzati sulla lunghezza del Segnale NVAL.
'
'   I vettori R(), X(), S() e F() sono dimensionati in questa routine.
'   Gli NN dati nei vettori sono organizzati come Vettore(0 To NN - 1).
'
'   Viene calcolata in R(0 To NFRE), X(0 To NFRE) solo la prima
'   meta' dello spettro per i valori di frequenza  NFRE = NVAL/2
'   piu' la Componente Continua in R(0).
'
'   Algoritmi tratti da: "Digital Time Series Analysis" di R. Otnes
'   e L. Enochson - pg. 138.  Nessun tentativo e' stato fatto per
'   migliorare la velocita' di esecuzione (tipo look-up table e/o
'   valutazione ricorsiva delle funzioni trigonometriche).
'
    Dim JF&, JD&, PI2_V#, Omega#, OmT#, NFRE_D#
'
    NFRE = Int(NVAL / 2)
    NFRE_D = CDbl(NVAL) / 2#
    PI2_V = PI2 / CDbl(NVAL)
'
    'ReDim R(0 To NVAL - 1), X(0 To NVAL - 1)    ' Per Spettro Completo.
    ReDim R(0 To NFRE), X(0 To NFRE)
    ReDim S(0 To NFRE), F(0 To NFRE)
'
    'For JF = 0 To NVAL - 1                      ' Per Spettro Completo.
    For JF = 0 To NFRE
        Omega = PI2_V * CDbl(JF)
        R(JF) = 0#
        X(JF) = 0#
'
        For JD = 0 To NVAL - 1
            OmT = Omega * CDbl(JD)
            R(JF) = R(JF) + D(JD) * Cos(OmT)
            X(JF) = X(JF) + D(JD) * Sin(OmT)
        Next JD
'
        ' Visualizzo (ed eventualmente interrompo)
        ' il progredire del calcolo:
        prbAvanz = 100 * JF / NFRE
        DoEvents
        If prbAvanz.Tag <> "" Then Exit Sub
    Next JF
'
    ' Calcolo dello Spettro di Ampiezza e delle Fasi:
    S(0) = Abs(R(0)) / CDbl(NVAL)
    F(0) = IIf(R(0) < 0#, PI, 0#)
    For JF = 1 To NFRE - 1
        S(JF) = Sqr(R(JF) * R(JF) + X(JF) * X(JF)) / NFRE_D
        F(JF) = DATAN2(-X(JF), R(JF))
    Next JF
    If (NVAL Mod 2) = 0 Then
        S(NFRE) = Abs(R(NFRE)) / CDbl(NVAL)
        F(NFRE) = IIf(R(NFRE) < 0#, PI, 0#)
    Else
        S(NFRE) = Sqr(R(NFRE) * R(NFRE) + X(NFRE) * X(NFRE)) / NFRE_D
        F(NFRE) = DATAN2(-X(NFRE), R(NFRE))
    End If
'
'
'
End Sub
Sub IFFT(R() As Double, X() As Double, Dr() As Double, Dx() As Double, _
    Optional ByRef NVAL&)
'
'   Definizioni:
'
'   IFFT = Inverse Fast Fourier Transform.
'
'   Entra con:
'       R()  = Vettore dei valori Reali      della Trasformata(f).
'       X()  = Vettore dei valori Immaginari della Trasformata(f).
'
'   Esce  con:
'       Dr() = Vettore dei dati Reali      del Segnale(t).
'       Dx() = Vettore dei dati Immaginari del Segnale(t).
'       NVAL = Numero  dei dati usati per l' antitrasformata.
'
'   I vettori Dr(), Dx() sono dimensionati in questa routine.
'   Gli NN dati nei vettori sono organizzati come Vettore(0 To NN - 1).
'
'   Algoritmo tratto da: "Theory and Application of Digital
'   Signal Processing" di L. Rabiner e B. Gold. - pg. 371.
'
    Dim NFRE&, M&, I&, NMN1&, J&
    Dim K&, L&, LE&, LE1&, IP&, JF&
    Dim Ur#, Ui#, Wr#, Wi#, Tr#, Ti#
'
    NVAL = UBound(R) + 1
    M = Int(Log(CDbl(NVAL) + 0.5) / Log2)
    NVAL = 2 ^ M
    NFRE = NVAL / 2
    NMN1 = NVAL - 1
'
    ReDim Dr(0 To NVAL - 1), Dx(0 To NVAL - 1), D(0 To NVAL - 1)
'
    ' Se non si vogliono sfruttare le proprieta'
    ' "transform-in-place" di questa implementazione:
    For I = 0 To NVAL - 1
        Dr(I) = R(I) / CDbl(NVAL)
        Dx(I) = -X(I) / CDbl(NVAL)
    Next I
'
    J = 0
    For I = 0 To NMN1 - 1
        If (I < J) Then
            Swap Dr(J), Dr(I)
            Swap Dx(J), Dx(I)
        End If
        K = NFRE
        Do While (K - 1 < J)
            J = J - K
            K = K / 2
        Loop
        J = J + K
    Next I
'
    For L = 1 To M
        LE = 2 ^ L
        LE1 = LE / 2
        Ur = 1#
        Ui = 0#
        Wr = Cos(PI / LE1)
        Wi = Sin(PI / LE1)
        For J = 0 To LE1 - 1
            For I = J To NVAL - 1 Step LE
                IP = I + LE1
                Tr = Dr(IP) * Ur - Dx(IP) * Ui
                Ti = Dr(IP) * Ui + Dx(IP) * Ur
                Dr(IP) = Dr(I) - Tr
                Dx(IP) = Dx(I) - Ti
                Dr(I) = Dr(I) + Tr
                Dx(I) = Dx(I) + Ti
            Next I
            Tr = Ur
            Ti = Ui
            Ur = Tr * Wr - Ti * Wi
            Ui = Tr * Wi + Ti * Wr
        Next J
    Next L
'
'
'
End Sub
Sub FFT(D() As Double, R() As Double, X() As Double, _
    S() As Double, F() As Double, ByRef NVAL&, ByRef NFRE&)
'
'   Definizioni:
'
'   FFT = Fast Fourier Transform.
'
'   Entra con:
'       D()   = Vettore dei dati Reali del Segnale(t).
'       NVAL  = Numero max. dei dati in D() da usare.
'
'   Esce  con:
'       R()   = Vettore dei valori Reali      della Trasformata(f).
'       X()   = Vettore dei valori Immaginari della Trasformata(f).
'       S()   = Vettore dello Spettro di Amp. della Trasformata(f).
'       F()   = Vettore delle Fasi            della Trasformata(f) [Radianti].
'       NVAL  = Numero  dei dati usati        dalla Trasformata(f).
'       NFRE  = Numero  dei valori di Frequenza calcolati.
'
'   I valori della Trasformata in R(), X() NON vengono
'   normalizzati sulla lunghezza del Segnale 2^M.
'
'   I vettori R(), X(), S() e F() sono dimensionati in questa routine.
'   Gli NN dati nei vettori sono organizzati come Vettore(0 To NN - 1).
'
'   Tradotta in Basic da: "Theory and Application of Digital
'   Signal Processing" di L. Rabiner e B. Gold. - pg. 367.
'
    Dim M&, I&, NMN1&, J&, K&, L&, LE&, LE1&, IP&, JF&
    Dim Ur#, Ui#, Wr#, Wi#, Tr#, Ti#
'
    M = Int(Log(CDbl(NVAL) + 0.5) / Log2)
    NVAL = 2 ^ M
    NFRE = NVAL / 2
    NMN1 = NVAL - 1
'
    ReDim R(0 To NVAL - 1), X(0 To NVAL - 1)
    ReDim S(0 To NFRE), F(0 To NFRE)
'
    ' Solo per segnali di ingresso reali e se non si vogliono
    ' sfruttare le proprieta' "transform-in-place" di questa
    ' implementazione:
    For I = 0 To NVAL - 1
        R(I) = D(I)
        X(I) = 0#           ' Per segnali di ingresso complessi
                            ' mettere la componente immaginaria.
    Next I
'
    J = 0
    For I = 0 To NMN1 - 1
        If (I < J) Then
            Swap R(J), R(I)
            'Swap X(J), X(I) ' Solo per segnali complessi.
        End If
        K = NFRE
        Do While (K - 1 < J)
            J = J - K
            K = K / 2
        Loop
        J = J + K
    Next I
'
    For L = 1 To M
        LE = 2 ^ L
        LE1 = LE / 2
        Ur = 1#
        Ui = 0#
        Wr = Cos(PI / LE1)
        Wi = Sin(PI / LE1)
        For J = 0 To LE1 - 1
            For I = J To NVAL - 1 Step LE
                IP = I + LE1
                Tr = R(IP) * Ur - X(IP) * Ui
                Ti = R(IP) * Ui + X(IP) * Ur
                R(IP) = R(I) - Tr
                X(IP) = X(I) - Ti
                R(I) = R(I) + Tr
                X(I) = X(I) + Ti
            Next I
            Tr = Ur
            Ti = Ui
            Ur = Tr * Wr - Ti * Wi
            Ui = Tr * Wi + Ti * Wr
        Next J
    Next L
'
    ' Calcolo dello Spettro di Ampiezza e delle Fasi:
    S(0) = Abs(R(0)) / CDbl(NVAL)
    F(0) = IIf(R(0) < 0#, PI, 0#)
    For JF = 1 To NFRE - 1
        S(JF) = Sqr(R(JF) * R(JF) + X(JF) * X(JF)) / CDbl(NFRE)
        F(JF) = DATAN2(-X(JF), R(JF))
    Next JF
    S(NFRE) = Abs(R(NFRE)) / CDbl(NVAL)
    F(NFRE) = IIf(R(NFRE) < 0#, PI, 0#)
'
'
'
End Sub
Public Function EliminaCC(D#(), Optional ByRef VMed#) As Double()
'
'   Ritorna i dati del vettore D() dopo l' eliminazione
'   della componente continua. I dati risultanti avranno,
'   quindi, valor medio = 0.  Ritorna anche VMed, valor
'   medio dei dati in D().
'
    Dim I&, dd#()
'
    dd() = D()
    VMed = 0#
'
    For I = 0 To UBound(dd)
        VMed = VMed + dd(I)
    Next I
    VMed = VMed / CDbl(UBound(dd) + 1)
'
    For I = 0 To UBound(dd)
        dd(I) = dd(I) - VMed
    Next I
'
    EliminaCC = dd()
'
'
'
End Function
Public Function ACorr(D#(), ByVal prbAvanz As ProgressBar, _
    Optional ByRef R2_I As Variant, _
    Optional ByVal NonCirc As Boolean = True, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione dell' autocorrelazione non circolare:
'                                  N-R
'   ACorr(R) = (1 / (N - R + 1)) * Som(D(I) * D(I + R))   con: N  = UBound(D)
'                                  I=0                     "   0 <= R2 <= N
'                                                          "   0 <= R  <= R2
'
'   Ritorna un vettore contenente la funzione di autocorrelazione
'   dei dati in D() calcolata su R2 ritardi [i.e. ACorr(0 To R2)].
'   Il segnale D() e' considerato finito e costituito da UBound(D) + 1
'   campioni.
'   Se il valore di R2 e' mancante viene usato UBound(D).
'   Se NonCirc = True i valori dell' autocorrelazione vengono
'   corretti per avere la non circolarita'.
'   Se NoCC = True dal segnale viene eliminata la componente continua.
'
'   Il calcolo viene effettuato per convoluzione diretta
'   e corretto con le equazioni di "Digital Time Series Analysis"
'   di R. K. Otnes, L. Enochson. - pg. 247.
'
    Dim I&, N&, R&, R2&, LNorm#, dd#()
'
    N = UBound(D)
    If IsMissing(R2_I) Then
        R2 = N
    Else
        R2 = MAX0(MIN0(CLng(R2_I), N), 0)
    End If
    R2_I = R2
'
    ReDim RR#(0 To R2)
    ReDim ACorr(0 To R2)
    LNorm = CDbl(N + 1)
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd() = EliminaCC(D())
    Else
        dd() = D()
    End If
'
    ' Calcolo la funzione di auto correlazione:
    For R = 0 To R2
        For I = 0 To N - R
            RR(R) = RR(R) + dd(I) * dd(I + R)
        Next I
        RR(R) = RR(R) / LNorm
'
        ' Visualizzo (ed eventualmente interrompo)
        ' il progredire del calcolo:
        prbAvanz = 100 * (R + 1) / (R2 + 1)
        DoEvents
        If prbAvanz.Tag <> "" Then Exit Function
    Next R
'
    If NonCirc Then
        ' Correzione dei valori dell' autocorrelazione
        ' per ottenere la non circolarita':
        For R = 0 To R2
            RR(R) = RR(R) * LNorm / CDbl(N - R + 1)
        Next R
    End If
'
    ACorr = RR()
'
'
'
End Function
Public Function ACorr_1(D#(), ByVal prbAvanz As ProgressBar, _
    Optional ByRef R2_I As Variant, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione dell' autocorrelazione:
'                                 N
'   ACorr_1(R) = (1 / (N + 1)) * Som(D(I) * D(I + R))   con: N  = UBound(D)
'                                I=0                     "   0 <= R2 <= N
'                                                        "   0 <= R  <= R2
'
'   Ritorna un vettore contenente la funzione di auto correlazione
'   dei dati in D() calcolata su R2 ritardi [i.e. ACorr_1(0 To R2)].
'   Il segnale D() e' considerato periodico con periodo pari a
'   UBound(D) + 1.
'   Se il valore di R2 e' mancante viene usato (UBound(D) + 1) \ 2.
'   Se NoCC = True dal segnale viene eliminata la componente continua.
'
'   Il calcolo viene effettuato per convoluzione diretta.
'
    Dim I&, N&, R&, R2&, dd#()
'
    N = UBound(D)
    If IsMissing(R2_I) Then
        R2 = (N + 1) \ 2
    Else
        R2 = MAX0(MIN0(CLng(R2_I), (N + 1) \ 2), 0)
    End If
    R2_I = R2
'
    ReDim RR#(0 To R2)
    ReDim ACorr_1(0 To R2)
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd() = EliminaCC(D())
    Else
        dd() = D()
    End If
    
'
    ' Completo il segnale DD() per renderlo
    ' periodico sull' intervallo considerato:
    ReDim Preserve dd(0 To N + R2)
    For I = 0 To R2 - 1
        dd(I + N + 1) = dd(I)
    Next I
'
    ' Calcolo la funzione di auto correlazione:
    For R = 0 To R2
        For I = 0 To N
            RR(R) = RR(R) + dd(I) * dd(I + R)
        Next I
        RR(R) = RR(R) / CDbl(N + 1)
'
        ' Visualizzo (ed eventualmente interrompo)
        ' il progredire del calcolo:
        prbAvanz = 100 * (R + 1) / (R2 + 1)
        DoEvents
        If prbAvanz.Tag <> "" Then Exit Function
    Next R
'
    ACorr_1 = RR()
'
'
'
End Function
Public Function MCorr(D1#(), D2#(), ByVal prbAvanz As ProgressBar, _
    Optional ByRef R1_I As Variant, Optional ByRef R2_I As Variant, _
    Optional ByVal NonCirc As Boolean = True, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione della mutua correlazione non circolare:
'                                    I2
'   MCorr(R) = (1 / (I2 - I1 + 1)) * Som(D1(I) * D2(I + R))   con:  N1  = UBound(D1)
'                                    I=I1                      "    N2  = UBound(D2)
'                                                              "   -N1 <= R1 <= 0
'                                                              "     0 <= R2 <= N2
'                                                              "    R1 <= R  <= R2
'                                                              "    I1  = MAX(0, -R)
'                                                              "    I2  = MIN(N1, N2 - R)
'
'   Ritorna un vettore contenente la funzione di mutua correlazione
'   fra i dati in D1() e quelli in D2() calcolata su R1 --> R2  ritardi
'   [i.e. MCorr(R1 To R2)].
'   Il segnale D2() e' considerato finito e costituito da UBound(D2) + 1
'   campioni.
'   Se il valore di R1 e' mancante viene usato -UBound(D1).
'   Se il valore di R2 e' mancante viene usato  UBound(D2).
'   Se NonCirc = True i valori della mutua correlazione vengono
'   corretti per avere la non circolarita'.
'   Se NoCC = True dai segnali viene eliminata la componente continua.
'
'   Il calcolo viene effettuato per convoluzione diretta
'   e corretto con le equazioni di "Digital Time Series Analysis"
'   di R. K. Otnes, L. Enochson. - pg. 247.
'
    Dim I&, I1&, I2&, N1&, N2&, R&, R1&, R2&
    Dim LNorm#, dd1#(), dd2#()
'
    N1 = UBound(D1)
    N2 = UBound(D2)
    If IsMissing(R1_I) Then
        R1 = -N1
    Else
        R1 = MIN0(MAX0(CLng(R1_I), -N1), 0)
    End If
    If IsMissing(R2_I) Then
        R2 = N2
    Else
        R2 = MAX0(MIN0(CLng(R2_I), N2), 0)
    End If
    R1_I = R1
    R2_I = R2
'
    ReDim RR#(R1 To R2)
    ReDim MCorr(R1 To R2)
    LNorm = CDbl(MIN0(N1 + 1, N2 + 1))
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd1() = EliminaCC(D1())
        dd2() = EliminaCC(D2())
    Else
    dd1() = D1()
    dd2() = D2()
    End If
'
    ' Calcolo la funzione di mutua correlazione:
    For R = R1 To R2
        I1 = MAX0(0, -R)
        I2 = MIN0(N1, N2 - R)
        For I = I1 To I2
            RR(R) = RR(R) + dd1(I) * dd2(I + R)
        Next I
        If NonCirc Then
            RR(R) = RR(R) / CDbl(I2 - I1 + 1)
        Else
            RR(R) = RR(R) / LNorm
        End If
'
        ' Visualizzo (ed eventualmente interrompo)
        ' il progredire del calcolo:
        prbAvanz = 100 * (R - R1 + 1) / (R2 - R1 + 1)
        DoEvents
        If prbAvanz.Tag <> "" Then Exit Function
    Next R
'
    MCorr = RR()
'
'
'
End Function
Public Function MCorr_1(D1#(), D2#(), ByVal prbAvanz As ProgressBar, _
    Optional ByRef R1_I As Variant, Optional ByRef R2_I As Variant, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione della mutua correlazione:
'                                 N1
'   MCorr_1(R) = (1 / (N1 + 1)) * Som(D1(I) * D2(I + R))   con:  N1  = UBound(D1)
'                                 I=0                       "    N2  = UBound(D2)
'                                                           "   -N1 <= R1 <= 0
'                                                           "     0 <= R2 <= N2
'                                                           "    R1 <= R  <= R2
'
'   Ritorna un vettore contenente la funzione di mutua correlazione
'   fra i dati in D1() e quelli in D2() calcolata su R1 --> R2  ritardi
'   [i.e. MCorr_1(R1 To R2)].
'   Il segnale D2() e' considerato periodico con periodo pari a
'   UBound(D2) + 1.
'   Il calcolo viene effettuato per convoluzione diretta.
'   Se il valore di R1 e' mancante viene usato -UBound(D1).
'   Se il valore di R2 e' mancante viene usato  UBound(D2).
'   Se NoCC = True dai segnali viene eliminata la componente continua.
'
'   Il calcolo viene effettuato per convoluzione diretta.
'
    Dim I&, N1&, N2&, R&, R1&, R2&, dd1#(), DD2_T#(), dd2#()
'
    N1 = UBound(D1)
    N2 = UBound(D2)
    If IsMissing(R1_I) Then
        R1 = -N1
    Else
        R1 = MIN0(MAX0(CLng(R1_I), -N1), 0)
    End If
    If IsMissing(R2_I) Then
        R2 = N2
    Else
        R2 = MAX0(MIN0(CLng(R2_I), N2), 0)
    End If
    R1_I = R1
    R2_I = R2
'
    ReDim RR#(R1 To R2)
    ReDim MCorr_1(R1 To R2)
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd1() = EliminaCC(D1())
        DD2_T() = EliminaCC(D2())
    Else
        dd1() = D1()
        DD2_T() = D2()
    End If
'
    ' Completo il segnale DD2() per renderlo
    ' periodico sull' intervallo considerato:
    ReDim dd2(R1 To N1 + R2)
    For I = R1 To N1 + R2
        dd2(I) = DD2_T(((I Mod (N2 + 1)) + N2 + 1) Mod (N2 + 1))
    Next I
'
    ' Calcolo la funzione di mutua correlazione:
    For R = R1 To R2
        For I = 0 To N1
            RR(R) = RR(R) + dd1(I) * dd2(I + R)
        Next I
        RR(R) = RR(R) / CDbl(N1 + 1)
'
        ' Visualizzo (ed eventualmente interrompo)
        ' il progredire del calcolo:
        prbAvanz = 100 * (R - R1 + 1) / (R2 - R1 + 1)
        DoEvents
        If prbAvanz.Tag <> "" Then Exit Function
    Next R
'
    MCorr_1 = RR()
'
'
'
End Function
Public Function MCorr_FT(D1#(), D2#(), _
    Optional ByRef R1_I As Variant, Optional ByRef R2_I As Variant, _
    Optional ByVal NonCirc As Boolean = True, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione della mutua correlazione non circolare:
'   MCorr_FT(R) = (L / Ld) * IFFT((1 / L) * Coniugata(FFT(D1(I))) * FFT(D2(I)))
'    con: N1  = UBound(D1),  N2 = UBound(D2),
'        -N1 <= R1 <= 0,  0 <= R2 <= N2,  R1 <= R <= R2,
'         L  = MIN(N1 + 1,  N2 + 1),  Ld = MIN(N1 + R, N2) - MAX(0, R) + 1
'
'   Ritorna un vettore contenente la funzione di mutua correlazione
'   fra i dati in D1() e quelli in D2() calcolata su R1 --> R2  ritardi
'   [i.e. MCorr_FT(R1 To R2)].
'   I segnali D1() e D2() sono considerati finiti.
'   Se il valore di R1 e' mancante viene usato -UBound(D1).
'   Se il valore di R2 e' mancante viene usato  UBound(D2).
'   Se NonCirc = True i valori della mutua correlazione vengono
'   corretti per avere la non circolarita'.
'   Se NoCC = True dai segnali viene eliminata la componente continua.
'
'   Il calcolo viene effettuato con FFT e IFFT.
'   Algoritmo tratto da: "Theory and Application of Digital
'   Signal Processing" di L. Rabiner e B. Gold. - pg. 403
'   e corretto con le equazioni di "Digital Time Series Analysis"
'   di R. K. Otnes, L. Enochson. - pg. 247, 248.
'
    Dim M&, N1&, N2&, R&, R1&, R2&, NFRE&
    Dim LNorm#, LNormD#, dd1#(), dd2#()
    Dim Re1#(), Im1#(), Re2#(), Im2#(), S#(), F#()
'
    N1 = UBound(D1)
    N2 = UBound(D2)
    If IsMissing(R1_I) Then
        R1 = -N1
    Else
        R1 = MIN0(MAX0(CLng(R1_I), -N1), 0)
    End If
    If IsMissing(R2_I) Then
        R2 = N2
    Else
        R2 = MAX0(MIN0(CLng(R2_I), N2), 0)
    End If
    R1_I = R1
    R2_I = R2
'
    M = Ceil(Log((N1 + N2 + 1)) / Log2)
    LNorm = CDbl(MIN0(N1 + 1, N2 + 1))
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd1() = EliminaCC(D1())
        dd2() = EliminaCC(D2())
    Else
        dd1() = D1()
        dd2() = D2()
    End If
'
    ReDim Preserve dd1(0 To (2 ^ M) - 1) ' Aggiungo zeri per avere una potenza di 2.
    ReDim Preserve dd2(0 To (2 ^ M) - 1) '    "      "    "    "    "     "    "  "
    ' Sposto i valori per comporre
    ' opportunamente il secondo segnale:
    For R = N2 To 0 Step -1
        dd2(N1 + R + 1) = dd2(R)
        dd2(R) = 0#
    Next R
'
    ' Calcolo la FFT dei due segnali:
    FFT dd1(), Re1(), Im1(), S(), F(), UBound(dd1) + 1, NFRE
    FFT dd2(), Re2(), Im2(), S(), F(), UBound(dd2) + 1, NFRE
'
    ' Moltiplico la seconda trasformata
    ' per la coniugata della prima:
    ReDim S(0 To UBound(Re1)), F(0 To UBound(Re1)) ' Usa S() e F() come vettori d' appoggio.
    For R = 0 To UBound(Re1)
        S(R) = (Re1(R) * Re2(R) + Im1(R) * Im2(R)) / LNorm  ' Parte reale.
        F(R) = (Re1(R) * Im2(R) - Re2(R) * Im1(R)) / LNorm  ' Parte immaginaria.
    Next R
'
    ' Calcolo l' antitrasformata corrispondente
    ' alla funzione di mutua correlazione:
    IFFT S(), F(), Re1(), Im1()     ' Usa Re1() e Im1() come vettori d' appoggio.
'
    ' Sistemo i valori nell' ordine dei ritardi:
    ReDim Im1(R1 To R2)             ' Usa Im1() come vettore d' appoggio.
    For R = R1 To R2
        Im1(R) = Re1(N1 + R + 1)    ' I valori > N1 + N2 + 1 sono nulli e/o simmetrici.
    Next R
'
    If NonCirc Then
        ' Correzione dei valori della mutua correlazione
        ' per ottenere la non circolarita':
        For R = R1 To R2
            ' LNormD = MIN(N1 + R, N2) - MAX(0, R) + 1:
            If N1 + R < N2 Then
                LNormD = CDbl(N1 + R + 1)
            Else
                LNormD = CDbl(N2 + 1)
            End If
            If R > 0 Then LNormD = LNormD - CDbl(R)
'
            Im1(R) = (LNorm / LNormD) * Im1(R)
        Next R
    End If
'
    MCorr_FT = Im1()
'
'
'
End Function
Public Function ACorr_FT(D#(), Optional ByRef R2_I As Variant, _
    Optional ByVal NonCirc As Boolean = True, _
    Optional ByVal NoCC As Boolean = True) As Double()
'
'   Espressione dell' autocorrelazione non circolare:
'   ACorr_FT(R) = ((N + 1) / (N - R + 1)) * IFFT((1 / (N + 1)) * |FFT(D(I))| ^ 2)
'    con: N = UBound(D),  0 <= R2 <= N,  0 <= R <= R2
'
'   Ritorna un vettore contenente la funzione di auto correlazione
'   dei dati in D() calcolata su R2 ritardi [i.e. ACorr_FT(0 To R2)].
'   Il segnale D() e' considerato finito e costituito da UBound(D) + 1
'   campioni.
'   Se il valore di R2 e' mancante viene usato UBound(D).
'   Se NonCirc = True i valori dell' autocorrelazione vengono
'   corretti per avere la non circolarita'.
'   Se NoCC = True dal segnale viene eliminata la componente continua.
'
'   Il calcolo viene effettuato con FFT e IFFT.
'   Algoritmo tratto da: "Theory and Application of Digital
'   Signal Processing" di L. Rabiner e B. Gold. - pg. 403
'   e corretto con le equazioni di "Digital Time Series Analysis"
'   di R. K. Otnes, L. Enochson. - pg. 247.
'
    Dim M&, N&, R&, R2&, NFRE&
    Dim LNorm#, dd#(), re#(), im#(), S#(), F#()
'
    N = UBound(D)
    If IsMissing(R2_I) Then
        R2 = N
    Else
        R2 = MAX0(MIN0(CLng(R2_I), N), 0)
    End If
    R2_I = R2
'
    M = Ceil(Log((2 * N + 1) + 0.5) / Log2)
    LNorm = CDbl(N + 1)
    ReDim ACorr_FT(0 To N)
'
    ' I dati devono avere media zero?
    If NoCC Then
        dd() = EliminaCC(D())
    Else
        dd() = D()
    End If
    ReDim Preserve dd(0 To (2 ^ M) - 1) ' Aggiungo zeri per avere una potenza di 2.
'
    ' Calcolo la FFT dei dati:
    FFT dd(), re(), im(), S(), F(), UBound(dd) + 1, NFRE
'
    ' Moltiplico la trasformata per la sua coniugata:
    ReDim S(0 To UBound(re)), F(0 To UBound(re)) ' Usa S() e F() come vettori d' appoggio.
    For R = 0 To UBound(re)
        S(R) = (re(R) * re(R) + im(R) * im(R)) / LNorm  ' Parte reale.
        F(R) = 0#                                       ' Parte immaginaria.
    Next R
'
    ' Calcolo l' antitrasformata corrispondente
    ' alla funzione di auto correlazione:
    IFFT S(), F(), re(), im()   ' Usa Re() e Im() come vettori d' appoggio.
    ReDim Preserve re(0 To R2)  ' Usa solo i primi R2 valori;
                                ' i valori > N sono nulli e/o simmetrici.
'
    If NonCirc Then
        ' Correzione dei valori dell' autocorrelazione
        ' per ottenere la non circolarita':
        For R = 0 To R2
            re(R) = (LNorm / CDbl(N - R + 1)) * re(R)
        Next R
    End If
'
    ACorr_FT = re()
'
'
'
End Function
Public Function MediaMobile(D#(), ByRef NsMM&) As Double()
'
'   Ritorna un vettore contenente la Media Mobile dei
'   dati in D() calcolata su NsMM campioni:
'
    Dim I&, VMed#
    ReDim dd#(0 To UBound(D))
'
    NsMM = MAX0(MIN0(NsMM, UBound(D) + 1), 1)
'
    For I = 0 To NsMM - 1
        VMed = VMed + D(I)
        dd(I) = VMed / CDbl(I + 1)
    Next I
'
    For I = NsMM To UBound(D)
        VMed = VMed + D(I) - D(I - NsMM)
        dd(I) = VMed / CDbl(NsMM)
    Next I
'
    MediaMobile = dd()
'
'
'
End Function


Public Function ValorEfficaceM(D#(), ByRef NsVEM&) As Double()
'
'   Ritorna un vettore contenente il valore efficace mobile
'   dei dati in D() calcolato su NsVEM campioni:
'
    Dim I&, VEff#
    ReDim dd#(0 To UBound(D))
'
    NsVEM = MAX0(MIN0(NsVEM, UBound(D) + 1), 1)
'
    For I = 0 To NsVEM - 1
        VEff = VEff + D(I) * D(I)
        dd(I) = Sqr(VEff / CDbl(I + 1))
    Next I
'
    For I = NsVEM To UBound(D)
        VEff = VEff + D(I) * D(I) - D(I - NsVEM) * D(I - NsVEM)
        dd(I) = Sqr(VEff / CDbl(NsVEM))
    Next I
'
    ValorEfficaceM = dd()
'
'
'
End Function
